"""Module for licensing and license serve checks."""

import logging
import os
import socket
import subprocess
import time

from ansys.mapdl.core.errors import LicenseServerConnectionError
from ansys.mapdl.core.misc import threaded_daemon

LOCALHOST = "127.0.0.1"
LIC_PATH_ENVAR = "ANSYSLIC_DIR"
LIC_FILE_ENVAR = "ANSYSLMD_LICENSE_FILE"
APP_NAME = "FEAT_ANSYS"  # TODO: We need to make sure this is the type of feature we need to checkout.
LIC_TO_CHECK = ["mech_1"]

LICENSES = {
    "ansys": "Ansys Mechanical Enterprise",
    "meba": "Ansys Mechanical Enterprise Solver",
    "mech_2": "Ansys Mechanical Premium",
    "mech_1": "Ansys Mechanical Pro",
}
ALLOWABLE_LICENSES = list(LICENSES)

## Regarding license checking.
# The available licenses we can check against are (in order of
# complete/comprehensiveness):
# 1. ``Ansys`` Enterprise license (the most complete)
# 2. ``meba`` Mechanical Enterprise license
# 3. ``mech_2`` Premium license
# 4. ``mech_1`` Pro license` (the most limited)
# To keep very general, and since we need just to be able to solve, we are going to check against the lower license (``mech_1``).

# TODO: Implement a warning for insufficient license rights.


LOG = logging.getLogger(__name__)
LOG.setLevel("CRITICAL")


def check_license_file(timeout=30, verbose=False):
    """Check the output of the license client log for connection error.

    Expect type of errors with 'DENIED' in the header such as:
    ```
    2021/09/06 22:39:38    DENIED              ansys                           21.2 (2021.0512)             1/0/0/0                 1/1/1/1   10268:FEAT_ANSYS:gayuso@AAPDDqVK5WqNLrt.win.ansys.com:winx64   7368:192.168.18.10
                Request name ansys does not exist in the licensing pool.
                Cannot connect to license server system.
                 The license server manager (lmgrd) has not been started yet,
                 the wrong port@host or license file is being used, or the
                 port or hostname in the license file has been changed.
                Feature:       ansys
                Server name:   192.168.18.10
                License path:  1055@AAPDDqVK5WqNLrt;
                FlexNet Licensing error:-15,578.  System Error: 10049 "WinSock: Invalid address"
    ```
    On the other side, successful license checkout will print `CHECKOUT       $APPNAME` in the file, such as:
    ```
    2021/09/20 12:59:49    CHECKOUT            FEAT_ANSYS                      21.2 (2021.0512)             1/1/1/1                 1/1/1/1   11704:FEAT_ANSYS:gayuso@AAPDDqVK5WqNLve.win.ansys.com:winx64   6884:192.168.56.1
    ```
    for `FEAT_ANSYS`.

    Parameters
    ----------
    timeout : int, optional
        Time to keep checking the license log file for errors. Default to 10 (seconds).

    Raises
    ------
    LicenseServerConnectionError
        If there is an error message in the license log file.
    TimeoutError
        Exceeded ``timeout`` while waiting for the license log file.

    """
    licdebug_file = os.path.join(get_licdebug_path(), get_licdebug_name())
    file_iterator = get_licdebug_tail(licdebug_file)

    max_time = time.time() + timeout
    while time.time() < max_time:
        msg = next(file_iterator)
        if msg:
            LOG.info(msg)
            if verbose:
                print(f"Output from {licdebug_file}:\n{msg}")

        if "DENIED" in msg:
            # read to the end of the file
            time.sleep(0.05)  # permit flush
            messages = [msg]
            while True:
                msg = next(file_iterator).strip()
                if not msg:
                    break
                messages.append(msg)

            raise LicenseServerConnectionError("\n".join(messages))

        if "CHECKOUT" in msg:
            # successful license checkout
            return True

    raise TimeoutError(
        f"Exceeded timeout of {timeout} seconds while examining:\n{licdebug_file}"
    )


def get_licdebug_path():
    """Get license client log (``licdebug``) path.

    This path is obtained from the correspondent env variable (OS
    dependent) and appending ``.ansys``.

    Returns
    -------
    str
        path of the license client log file.
    """
    if os.name == "nt":
        folder = os.getenv("TEMP")
    elif os.name == "posix":
        folder = os.getenv("HOME")
    else:
        raise OSError(f"Unsupported OS {os.name}")

    return os.path.join(folder, ".ansys")


def get_licdebug_name():
    """Get license client log file name.

    This file change the name according to the ANSYS version and the type of license requested (``$appname``).
    * For ANSYS version 22.1 and above: ``licdebug.$hostname.$appname.$version.out``
    * For ANSYS version 21.2 and below: ``licdebug.$appname.$version.out``

    where:
    * ``$hostname`` is the name of the machine.
    * ``$appname`` is the name of the feature used by the license client. Eg. 'FEAT_ANSYS'
    * ``$version`` is the version of ANSYS. Eg 211 for version 21.1.

    Returns
    -------
    str
        licdebug log file complete name.

    """
    # Licdebug name convention:
    # - For version 22.1 and above: `licdebug.$hostname.$appname.$version.out`
    # - For version 21.2 and below: `licdebug.$appname.$version.out`

    from ansys.mapdl.core.launcher import _version_from_path, get_ansys_path

    name = "licdebug"
    hostname = socket.gethostname()
    appname = APP_NAME
    # This is the type of license my client requests (Windows 10, 2021R2)
    version = _version_from_path(get_ansys_path(allow_input=False))
    ending = "out"

    if version < 221:
        parts = (name, appname, version, ending)
    else:
        parts = (name, hostname, appname, version, ending)

    return ".".join([str(each_part) for each_part in parts])


def get_licdebug_tail(licdebug_file, start_timeout=10):
    """Get each of the licdebug file messages.

    This method keeps the ``licdebug`` file open checking for complete messages.
    It yields one message at a time when called.

    Parameters
    ----------
    licdebug_file : str
        Path to the ``licdebug`` file.wh
    start_timeout : float, optional
        Maximum timeout to wait until the file exists.

    Yields
    ------
    msg : str
        Message formatted as a single string.

    """
    # wait until file exists
    max_time = time.time() + start_timeout
    while not os.path.isfile(licdebug_file):
        time.sleep(0.01)
        if time.time() > max_time:
            raise TimeoutError(
                f"Exceeded {start_timeout} seconds while waiting for {licdebug_file}"
                " to exist."
            )

    with open(licdebug_file) as fid:
        # Going to the end of the file.
        fid.seek(0, 2)
        while True:
            lines = "".join(fid.readlines())
            yield lines


def get_ansyslic_dir():
    """Get the path to the Ansys license directory"""

    # it's possible the user has specified the license as an env var
    ansyslic_dir = None
    if LIC_FILE_ENVAR in os.environ:
        ansyslmd_var = os.environ[LIC_FILE_ENVAR]
        if not os.path.isfile(ansyslmd_var):
            # likely license info
            ansyslic_dir = None
    else:
        ansyslic_dir = os.getenv(LIC_PATH_ENVAR)

    # env var may not be specified, check in the usual location
    if ansyslic_dir is None:
        if os.name == "nt":
            ansyslic_dir = os.path.join(
                os.environ["ProgramFiles"],
                "ANSYS Inc",
                "Shared Files",
                "Licensing",
            )
        else:
            ansyslic_dir = "/usr/ansys_inc/shared_files/licensing"

        if not os.path.isdir(ansyslic_dir):
            raise FileNotFoundError(
                f"Unable to locate ANSYS licencing path at {ansyslic_dir}\n"
                f"Specify the {LIC_PATH_ENVAR}"
            )

    return ansyslic_dir


def check_mech_license_available(host=None, verbose=False, licenses=None):
    """Check if there mechanical license available by running 'ansysli_util'.

    This uses the default configuration available to MAPDL.

    .. warning::
       This method to check the license server status is not stable.  It
       is very likely it will show license not available when it actually
       is.

    Parameters
    ----------
    host : str, optional
        Override the default license server.  By default, use the
        values available in the Ansys license file.
    verbose : bool, optional
        Print output while checking for the license.
    license : str, list, optional
        A list or single license to check for.  One or more of the
        following:

        1. ``"Ansys"`` Enterprise license (the most complete)
        2. ``"meba"`` Mechanical Enterprise license
        3. ``"mech_2"`` Premium license
        4. ``"mech_1"`` Pro license` (the most limited)

    Returns
    -------
    bool
        Returns True when the license exists.

    Raises
    ------
    LicenseServerConnectionError
        When errors messages found in the output of the license file.

    """

    if licenses is None:
        licenses = LIC_TO_CHECK
    elif isinstance(licenses, str):
        licenses = [licenses]

    msg1 = "No such feature exists"
    msg2 = "The server is down or is not responsive."
    for each_license in licenses:
        output = checkout_license(each_license, host, verbose=verbose)
        if msg1 in output or msg2 in output:
            raise LicenseServerConnectionError(output)

    return True


def checkout_license(lic, host=None, port=2325, verbose=False):
    """Check if a license is available using the Ansys license utility.

    It uses it own process.

    Parameters
    ----------
    lic : str
        License type.  For example, ``"meba".  one of the following:

        1. ``"Ansys"`` Enterprise license (the most complete)
        2. ``"meba"`` Mechanical Enterprise license
        3. ``"mech_2"`` Premium license
        4. ``"mech_1"`` Pro license` (the most limited)
    host : str, optional
        Host to attempt to checkout license from.  When set, this
        overrides any settings from the default license path.
    port : int, optional
        Port on the host to connect to.  Only used when ``host`` is set.

    """
    if lic.lower() not in ALLOWABLE_LICENSES:
        raise ValueError(f"Invalid license '{lic}'")

    ansyslic_dir = get_ansyslic_dir()
    if os.name == "nt":
        ansysli_util_path = os.path.join(ansyslic_dir, "winx64", "ansysli_util.exe")
    else:
        ansysli_util_path = os.path.join(ansyslic_dir, "linx64", "ansysli_util")

    if not os.path.isfile(ansysli_util_path):
        raise FileNotFoundError(
            "Ansys licensing path exists but ansysli_util not found at:\n"
            f"{ansysli_util_path}"
        )

    # allow the specification of ip and port
    env = os.environ.copy()
    if host is not None and port is not None:
        env["ANSYSLI_SERVERS"] = f"{host}:{port}"
        env["ANS_FLEXLM_DISABLE_DEFLICPATH"] = "TRUE"

    tstart = time.time()
    process = subprocess.Popen(
        f'"{ansysli_util_path}" -checkout {lic}',
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        env=env,
        shell=True,
    )
    output = process.stdout.read().decode()
    if verbose:
        t_elap = time.time() - tstart
        print(f"License check complete in {t_elap:.2} seconds.\n")
        print(output)
    return output


class LicenseChecker:
    """License checker class.

    Two methods are used and exposed with the :func:`LicenseChecker.start` method:

    * Check the ``licdebug`` log file for errors.
    * Check the available mechanical licenses using ``ansysli_util`` executable.

    Parameters
    ----------
    timeout : float, optional
        Timeout for the licensing log file check.
    verbose : bool, optional
        Enable or disable verbose output of the license check.  Useful
        for debugging.

    """

    def __init__(self, timeout=30, verbose=False):
        self._license_file_msg = []
        self._license_file_success = None

        self._license_checkout_msg = []
        self._license_checkout_success = None
        self._timeout = timeout
        self._verbose = verbose

        self._lic_file_thread = None
        self._checkout_thread = None

    @threaded_daemon
    def check_license_file(self):
        try:
            check_license_file(self._timeout, verbose=self._verbose)
        except Exception as error:
            self._license_file_success = False
            self._license_file_msg.append(str(error))
        else:
            self._license_file_success = True

    @threaded_daemon
    def checkout_license(self, host=None):
        try:
            check_mech_license_available(host, verbose=self._verbose)
        except Exception as error:
            self._license_checkout_success = False
            self._license_checkout_msg.append(str(error))
        else:
            self._license_checkout_success = True

    def start(self, license_file=True, checkout_license=False):
        """Start monitoring the license file and attempt a license checkout.

        Parameters
        ----------
        license_file : bool, optional
            Start the license file thread.
        checkout_license : bool, optional
            Start the checkout license thread.  By default this is
            disabled.

        """
        if license_file:
            self._lic_file_thread = self.check_license_file()
        if checkout_license:
            self._checkout_thread = self.checkout_license()

    def wait(self):
        """Wait until the license checks are complete or have timed out."""
        if self._lic_file_thread is not None:
            self._lic_file_thread.join()
        if self._checkout_thread is not None:
            self._checkout_thread.join()

    def check(self):
        """Report if the license checkout or license check was successful.

        It first checks the output from the license file and later the
        output from the checkout process.

        Returns
        -------
        bool
            ``True`` When license successfully checked out, ``False``
            when license check failed and nothing to report.  Checkout
            failure will raise
            :class:`ansys.mapdl.core.errors.LicenseServerConnectionError``.

        Raises
        ------
        LicenseServerConnectionError
            If there were any errors during the license checkout or
            license file check.

        """
        if self._license_file_success:
            return True
        elif self._license_file_success is False:
            raise LicenseServerConnectionError("\n".join(self._license_file_msg))

        if self._license_checkout_success:
            return True
        elif self._license_checkout_success is False:
            raise LicenseServerConnectionError("\n".join(self._license_checkout_msg))

        return False



class MagneticsCalc:
    def curr2d(self, **kwargs):
        """Calculates current flow in a 2-D conductor.

        APDL Command: CURR2D

        Notes
        -----
        CURR2D invokes an ANSYS macro which calculates the total current
        flowing in a conducting body for a 2-D planar or axisymmetric magnetic
        field analysis.  The currents may be applied source currents or induced
        currents (eddy currents).  The elements of the conducting region must
        be selected before this command is issued.  The total current
        calculated by the macro is stored in the parameter TCURR.  Also, the
        total current and total current density are stored on a per-element
        basis in the element table [ETABLE] with the labels TCURR and JT,
        respectively.  Use the PLETAB and PRETAB commands to plot and list the
        element table items.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"CURR2D,"
        return self.run(command, **kwargs)

    def emagerr(self, **kwargs):
        """Calculates the relative error in an electrostatic or electromagnetic

        APDL Command: EMAGERR
        field analysis.

        Notes
        -----
        The relative error is an approximation of the mesh discretization error
        associated with a solution. It is based on the discrepancy between the
        unaveraged, element-nodal field values and the averaged, nodal field
        values. The calculation is valid within a material boundary and does
        not consider the error in continuity of fields across dissimilar
        materials.

        For electrostatics, the field values evaluated are the electric field
        strength (EFSUM) and the electric flux density (DSUM). A relative error
        norm of each is calculated on a per-element basis and stored in the
        element table [ETABLE] with the labels EF_ERR and D_ERR. Normalized
        error values EFN_ERR and DN_ERR are also calculated and stored in the
        element table. Corresponding quantities for electromagnetics are H_ERR,
        B_ERR, HN_ERR, and BN_ERR, which are calculated from the magnetic field
        intensity (HSUM) and the magnetic flux density (BSUM).  The normalized
        error value is the relative error norm value divided by the peak
        element-nodal field value for the currently selected elements.

        Use the PLETAB and PRETAB commands to plot and list the error norms and
        normalized error values.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"EMAGERR,"
        return self.run(command, **kwargs)

    def emf(self, **kwargs):
        """Calculates the electromotive force (emf), or voltage drop along a

        APDL Command: EMF
        predefined path.

        Notes
        -----
        EMF invokes an ANSYS macro which calculates the electromotive force
        (emf), or voltage drop along a predefined path (specified with the PATH
        command). It is valid for both 2-D and 3-D electric field analysis or
        high-frequency electromagnetic field analysis. The calculated emf value
        is stored in the parameter EMF.

        You must define a line path (via the PATH command) before issuing the
        EMF command macro. The macro uses calculated values of the electric
        field (EF), and uses path operations for the calculations. All path
        items are cleared when the macro finishes executing.

        The EMF macro sets the "ACCURATE" mapping method and "MAT"
        discontinuity option on the PMAP command. The ANSYS program retains
        these settings after executing the macro.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"EMF,"
        return self.run(command, **kwargs)

    def emft(self, **kwargs):
        """Summarizes electromagnetic forces and torques.

        APDL Command: EMFT

        Notes
        -----
        Use this command to summarize electromagnetic force and torque in both
        static electric and magnetic problems. To use this command, select the
        nodes in the region of interest and make sure that all elements are
        selected. If RSYS = 0, the force is reported in the global Cartesian
        coordinate system. If RSYS ≠ 0, force is reported in the specified
        coordinate system. However, for torque, if RSYS ≠ 0, this command will
        account for the shift and rotation as specified by RSYS, but will
        report only the Cartesian components.

        Forces are stored as items _FXSUM, _FYSUM, _FZSUM, and _FSSUM. Torque
        is stored as items _TXSUM, _TYSUM, _TZSUM, and _TSSUM.

        This command is valid only with PLANE121, SOLID122, SOLID123, PLANE233,
        SOLID236 and SOLID237 elements. For any other elements, you must use
        FMAGSUM.
        """
        command = f"EMFT,"
        return self.run(command, **kwargs)

    def fluxv(self, **kwargs):
        """Calculates the flux passing through a closed contour.

        APDL Command: FLUXV

        Notes
        -----
        FLUXV invokes an ANSYS macro which calculates the flux passing through
        a closed contour (path) predefined by PATH.  The calculated flux is
        stored in the parameter FLUX.  In a 2-D analysis, at least two nodes
        must be defined on the path.  In 3-D, a path of nodes describing a
        closed contour must be specified (i.e., the first and last node in the
        path specification must be the same).  A counterclockwise ordering of
        nodes on the PPATH command will give the correct sign on flux.  Path
        operations are used for the calculations, and all path items are
        cleared upon completion.  This macro is only available for vector
        potential formulations.
        """
        command = f"FLUXV,"
        return self.run(command, **kwargs)

    def mmf(self, **kwargs):
        """Calculates the magnetomotive force along a path.

        APDL Command: MMF

        Notes
        -----
        MMF invokes an ANSYS macro which calculates the magnetomotive force
        (mmf) along a predefined path [PATH].  It  is valid for both 2-D and
        3-D magnetic field analyses.  The calculated mmf value is stored in the
        parameter MMF.

        A closed path [PATH], passing through the magnetic circuit for which
        mmf is to be calculated, must be defined before this command is issued.
        A counterclockwise ordering of points on the PPATH command will yield
        the correct sign on the mmf.  The mmf is based on Ampere's Law.  The
        macro makes use of calculated values of field intensity (H), and uses
        path operations for the calculations.  All path items are cleared upon
        completion.  The MMF macro sets the "ACCURATE" mapping method and "MAT"
        discontinuity option of the PMAP command.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MMF,"
        return self.run(command, **kwargs)

    def plf2d(self, ncont="", olay="", anum="", win="", **kwargs):
        """Generates a contour line plot of equipotentials.

        APDL Command: PLF2D

        Parameters
        ----------
        ncont
            Number of contour lines to display.  Issue in multiples of 9 (i.e.,
            9, 18, 27, etc.).  Default is 27 contour lines.

        olay
            Overlay:

            0 - Overlay edge outlines by material number.

            1 - Overlay edge outlines by real constant number.

        anum
            Highest material or real constant attribute number.  Command will
            cycle through ANUM element display overlays.  Defaults to 10.

        win
            Window number to which command applies.  Defaults to 1.

        Notes
        -----
        PLF2D invokes an ANSYS macro which plots equipotentials of the degree
        of freedom AZ.  These equipotential lines are parallel to flux lines
        and thus give a good representation of flux patterns.  In the
        axisymmetric case, the display is actually ``r*AZ`` where "r" is the node
        radius.  The macro overlays (OLAY) edge outlines by material number or
        real constant number (ANUM) and allows user control over the number of
        contour lines to display (NCONT).

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"PLF2D,{ncont},{olay},{anum},{win}"
        return self.run(command, **kwargs)

    def powerh(self, **kwargs):
        """Calculates the rms power loss in a conductor or lossy dielectric.

        APDL Command: POWERH

        Notes
        -----
        POWERH invokes an ANSYS macro which calculates the time-averaged (rms)
        power loss in a conductor or lossy dielectric material from a harmonic
        analysis.  The power loss is stored in the parameter PAVG.  Conductor
        losses include solid conductors and surface conductors approximated by
        impedance or shielding boundary conditions. The power loss density for
        solid conductors or dielectrics is stored in the element table with the
        label PLOSSD and may be listed [PRETAB] or displayed [PLETAB].  PLOSSD
        does not include surface losses. The elements of the conducting region
        must be selected before this command is issued.  POWERH is valid for
        2-D and 3-D analyses.
        """
        command = f"POWERH,"
        return self.run(command, **kwargs)

    def senergy(self, opt="", antype="", **kwargs):
        """Determines the stored magnetic energy or co-energy.

        APDL Command: SENERGY

        Parameters
        ----------
        opt
            Item to be calculated:

            0 - Stored magnetic energy.

            1 - Stored magnetic co-energy.

        antype
            Analysis type:

            0 - Static or transient.

            1 - Harmonic.

        Notes
        -----
        SENERGY invokes an ANSYS macro which calculates the stored magnetic
        energy or co-energy for all selected elements.  (For a harmonic
        analysis, the macro calculates a time-averaged (rms) stored energy.)  A
        summary table listing the energy by material number is produced.  The
        energy density is also calculated and stored on a per-element basis in
        the element table [ETABLE] with the label MG_ENG (energy density) or
        MG_COENG (co-energy density).  The macro erases all other items in the
        element table [ETABLE] and only retains the energy density or co-energy
        density.  Use the PLETAB and PRETAB commands to plot and list the
        energy density.  The macro is valid for static and low-frequency
        magnetic field formulations.  The macro will not calculate stored
        energy and co-energy for the following cases:

        Orthotropic nonlinear permanent magnets.

        Orthotropic nonlinear permeable materials.

        Temperature dependent materials.

        SENERGY is restricted to MKSA units.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"SENERGY,{opt},{antype}"
        return self.run(command, **kwargs)



"""
Implements a simple version check for a client/server pair.

Used when you want to verify if your server version is a minimum
value.

The decorator allows arguments within the decorator itself.

"""

# Map version tuple to MAPDL release version
VERSION_MAP = {
    (0, 0, 0): "2020R2",
    (0, 3, 0): "2021R1",
    (0, 4, 0): "2021R2",
    (0, 4, 1): "2021R2",
    (0, 5, 0): "2022R1",
    (0, 5, 1): "2022R2",  # as of 21 Mar 2022 unreleased
}


def meets_version(version, meets):
    """Check if a version string meets a minimum version.
    This is a simplified way to compare version strings. For a more robust
    tool, please check out the ``packaging`` library:
    https://github.com/pypa/packaging

    Parameters
    ----------
    version : str
        Version string.  For example ``'0.25.1'``.

    meets : str
        Version string.  For example ``'0.25.2'``.

    Returns
    -------
    newer : bool
        True if version ``version`` is greater or equal to version ``meets``.

    Examples
    --------
    >>> meets_version('0.25.1', '0.25.2')
    False
    >>> meets_version('0.26.0', '0.25.2')
    True
    """
    if not isinstance(version, tuple):
        va = version_tuple(version)
    else:
        va = version

    if not isinstance(meets, tuple):
        vb = version_tuple(meets)
    else:
        vb = meets

    if len(va) != len(vb):
        raise ValueError("Versions are not comparable.")

    for i in range(len(va)):
        if va[i] > vb[i]:
            return True
        elif va[i] < vb[i]:
            return False

    # Arrived here if same version
    return True


def version_tuple(v):
    """Convert a version string to a tuple containing ints.
    Non-numeric version strings will be converted to 0.  For example:
    ``'0.28.0dev0'`` will be converted to ``'0.28.0'``
    Returns
    -------
    ver_tuple : tuple
        Length 3 tuple representing the major, minor, and patch
        version.
    """
    split_v = v.split(".")
    while len(split_v) < 3:
        split_v.append("0")

    if len(split_v) > 3:
        raise ValueError(
            "Version strings containing more than three parts " "cannot be parsed"
        )

    vals = []
    for item in split_v:
        if item.isnumeric():
            vals.append(int(item))
        else:
            vals.append(0)

    return tuple(vals)


class VersionError(ValueError):
    """Raised when the Server is the wrong version"""

    def __init__(self, msg="Invalid Server version"):
        ValueError.__init__(self, msg)


def version_requires(min_version):
    """Ensure the method called matches a certain version.

    Example usage:

    class Client():

        def __init__(self):
            '''Connects to a fake server'''
            self._server = FakeServer()

        @version_requires((0, 1, 3))  # require 0.1.3
        def meth_a(self):
            '''calls method a on the 'server''''
            return self._server.meth_a()

        @version_requires((0, 2, 3))  # require 0.2.3
        def meth_b(self):
            '''calls method b on the 'server''''
            return self._server.meth_b()

    """

    def decorator(func):
        # first arg *must* be a tuple containing the version
        if not isinstance(min_version, tuple):
            raise TypeError(
                "version_requires decorator must include a version "
                "tuple.  For example:\n"
                "``@_version_requires((0, 1, 3))``"
            )
        if not len(min_version) == 3:
            raise TypeError(
                "version_requires decorator must include a version "
                "tuple.  For example:\n"
                "``@_version_requires((0, 1, 3))``"
            )

        def wrapper(self, *args, **kwargs):
            """Call the original function"""
            # must be called from a "Client" instance containing a server attribute
            if not hasattr(self, "_server_version"):
                raise AttributeError(
                    "decorated class must have `_server_version` " "attribute"
                )

            if not meets_version(self._server_version, min_version):

                # try to give the user a helpful warning indicating
                # the minimum version of MAPDL
                if min_version in VERSION_MAP:
                    raise VersionError(
                        f"``{func.__name__}`` requires MAPDL version "
                        f">= {VERSION_MAP[min_version]}"
                    )

                # otherwise, use the less helpful "gRPC server" version
                raise VersionError(
                    f"``{func.__name__}`` requires gRPC server "
                    f"version >= {min_version}"
                )

            return func(self, *args, **kwargs)

        return wrapper

    return decorator



class MultiFieldSolverLoadTransfer:
    def mflcomm(
        self,
        type_="",
        fname1="",
        intname1="",
        label1="",
        fname2="",
        intname2="",
        label2="",
        option="",
        **kwargs,
    ):
        """Defines a load transfer for code coupling analyses.

        APDL Command: MFLCOMM

        Parameters
        ----------
        type\_
            Set to SURF for a surface load transfer. Only surface load
            transfers are available for MFX.

        fname1
            Sets the field solver name for the server (sending) code with a
            case-sensitive character string of up to 80 characters.

        intname1
            Sets the interface name or number for the field solver of the
            server code. ANSYS interfaces are numbered and are defined by the
            SF family of commands (SF, SFA, or SFE) with the FSIN surface load
            label. CFX interfaces use names, which are set in CFX-Pre.

        label1
            Sets the surface load label for the field solver of the server code
            with a character string of up to 80 characters. ANSYS uses a
            combination of the label and option to determine what data is
            transferred (e.g., heat flows and not fluxes are sent with the
            label/option pair HFLU/CPP). ANSYS cannot serve total force or
            total force density to CFX for either formulation. CFX will send
            the data requested by the label regardless of the option. CFX
            labels that have more than one word must be enclosed in single
            quotes. Note that this field is case-sensitive; i.e., FORC will
            work, but forc will not.

        fname2
            Sets the field solver name for the client (receiving) code with a
            character string of up to 80 characters.

        intname2
            Sets the interface name or number for the field solver of the
            client code with a character string of up to 80 characters. ANSYS
            interfaces are numbered and are defined by the SF family of
            commands (SF, SFA, or SFE) with the FSIN surface load label. CFX
            interfaces use names, which are set in CFX-Pre.

        label2
            Sets the surface load label for the field solver of the client code
            with a character string of up to 80 characters. ANSYS uses a
            combination of the label and option to determine what data is
            transferred (e.g., heat flows and not fluxes are sent with the
            label-option pair HFLU/CPP). CFX will send the data requested by
            the label regardless of the option. CFX labels that have more than
            one word must be enclosed in single quotes. Note that this field is
            case-sensitive; i.e., FORC will work, but forc will not.

        option
            NONC

            NONC - Profile preserving: Sets the interface load transfer to the nonconservative
                   formulation (default for displacement and temperature). In
                   the nonconservative formulation, the force density (or heat
                   flux) is transferred across the interface, preserving the
                   density profile between the two fields.

            CPP - Conservative: Uses a local conservative formulation while preserving the
                  density profile (default for total force and wall heat flow).
                  In the conservative formulation, total force (or heat flow)
                  must be transferred across the interface from the CFX field
                  solver to the ANSYS field solver.

        Notes
        -----
        ANSYS input should always be in consistent units for its model.

        ANSYS uses a combination of the label and option to determine what data
        to transfer. CFX will send exactly the data requested by the label,
        regardless of the option. However, for the NONC option, the CFX label
        must be Total Force Density or Wall Heat Flux and for the CPP option,
        the CFX label must be Total Force or Wall Heat Flow.

        For more information on profile preserving and conservative load
        transfer, see Load Interpolation in the Coupled-Field Analysis Guide.
        Mapping Diagnostics are also available; however, if the improperly-
        mapped nodes are based on the CFX mesh, you should ignore the ANSYS-
        generated components because the CFX nodes are not present in the ANSYS
        database.

        If you are working interactively, you can choose two pre-defined
        combinations, Mechanical or Thermal, or you can choose a Custom option.
        If you choose the Mechanical load type, then the Total Force Density
        and Total Mesh Displacement data (corresponding to the ANSYS FORC and
        DISP labels, respectively) is transferred. If you choose the Thermal
        load type, then the Temperature and Wall Heat Flux data (corresponding
        to the ANSYS TEMP and HFLU labels, respectively) is transferred. If you
        choose Custom, you can select any valid combination of label and option
        as described above.

        The ANSYS Multi-field solver solver does not allow you to switch the
        load transfer direction for the same load quantity across the same
        interfaces for a restart run. For example, if Field1 sends temperature
        to and receives heat flow from Field2 across Interface 1 in a previous
        solution, then you cannot make Field1 send heat flow to and receive
        temperatures from Field2 across the same interface in a restart run,
        even if you cleared the corresponding load transfer command.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFLCOMM,{type_},{fname1},{intname1},{label1},{fname2},{intname2},{label2},{option}"
        return self.run(command, **kwargs)

    def mfsurface(self, inumb="", fnumb1="", label="", fnumb2="", **kwargs):
        """Defines a surface load transfer for an ANSYS Multi-field solver

        APDL Command: MFSURFACE
        analysis.

        Parameters
        ----------
        inumb
            Interface number for load transfer. The interface number
            corresponds to the interface number specified by the surface flag
            FSIN (SFxxcommands).

        fnumb1
            Field number of sending field.

        label
            Valid surface load labels:

        fnumb2
            Field number for receiving field.

        Notes
        -----
        This command is also valid in PREP7.

        The ANSYS Multi-field solver solver does not allow you to switch the
        load transfer direction for the same load quantity across the same
        interfaces for a restart run. For example, if Field1 sends temperature
        to and receives heat flow from Field2 across Interface 1 in a previous
        solution, then you cannot make Field1 send heat flow to and receive
        temperatures from Field2 across the same interface in a restart run,
        even if you cleared the corresponding load transfer command.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFSURFACE,{inumb},{fnumb1},{label},{fnumb2}"
        return self.run(command, **kwargs)

    def mfvolume(self, inumb="", fnumb1="", label="", fnumb2="", **kwargs):
        """Defines a volume load transfer for an ANSYS Multi-field solver

        APDL Command: MFVOLUME
        analysis.

        Parameters
        ----------
        inumb
            Interface number for load transfer. The interface number
            corresponds to the interface number specified by the volume flag
            FVIN (BFE command).

        fnumb1
            Field number of sending field.

        label
            Valid volume load labels:

        fnumb2
            Field number for receiving field.

        Notes
        -----
        This command is also valid in PREP7.

        The ANSYS Multi-field solver solver does not allow you to switch the
        load transfer direction for the same load quantity across the same
        interfaces for a restart run. For example, if Field1 sends temperature
        to and receives heat flow from Field2 across Interface 1 in a previous
        solution, then you cannot make Field1 send heat flow to and receive
        temperatures from Field2 across the same interface in a restart run,
        even if you cleared the corresponding load transfer command.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFVOLUME,{inumb},{fnumb1},{label},{fnumb2}"
        return self.run(command, **kwargs)



"""Module to support MAPDL CAD geometry"""
import re

import numpy as np

from ansys.mapdl.core import _HAS_PYVISTA

if _HAS_PYVISTA:
    import pyvista as pv

from ansys.mapdl.core.misc import run_as_prep7, supress_logging

VALID_TYPE_MSG = """- 'S' : Select a new set (default)
- 'R' : Reselect a set from the current set.
- 'A' : Additionally select a set and extend the current set.
- 'U' : Unselect a set from the current set.
"""

FLST_LOOKUP = {
    "NODE": 1,  # node numbers
    "ELEM": 2,  # element numbers
    "KP": 3,  # keypoint numbers
    "LINE": 4,  # line numbers
    "AREA": 5,  # area numbers
    "VOLU": 6,  # volume numbers
    "TRACE": 7,  # trace points
    "COORD": 8,  # coordinate locations
}


def merge_polydata(items):
    """Merge list of polydata or unstructured grids"""

    # lazy import here for faster module loading
    try:
        from pyvista._vtk import vtkAppendPolyData
    except:
        from vtk import vtkAppendPolyData

    afilter = vtkAppendPolyData()
    for item in items:
        afilter.AddInputData(item)
        afilter.Update()

    return pv.wrap(afilter.GetOutput())


def get_elements_per_area(resp):
    """Get the number of elements meshed for each area given the response
    from ``AMESH``.

    GENERATE NODES AND ELEMENTS   IN  ALL  SELECTED AREAS
        ** AREA     1 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA     2 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA     3 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA     4 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA     5 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA     6 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA     7 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA     8 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA     9 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA    10 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA    11 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **
        ** AREA    12 MESHED WITH      64 QUADRILATERALS,        0 TRIANGLES **

     NUMBER OF AREAS MESHED     =         12
     MAXIMUM NODE NUMBER        =        772
     MAXIMUM ELEMENT NUMBER     =        768

    Returns
    -------
    list
        List of tuples, each containing the area number and number of
        elements per area.

    """
    # MAPDL changed their output at some point.  Check for both output types.
    reg = re.compile(r"Meshing of area (\d*) completed \*\* (\d*) elements")
    groups = reg.findall(resp)
    if groups:
        groups = [[int(anum), int(nelem)] for anum, nelem in groups]
    else:
        reg = re.compile(r"AREA\s*(\d*).*?(\d*)\s*QUADRILATERALS,\s*(\d*) TRIANGLES")
        groups = reg.findall(resp)
        groups = [(int(anum), int(nquad) + int(ntri)) for anum, nquad, ntri in groups]

    return groups


class Geometry:
    """Pythonic representation of MAPDL CAD geometry

    Contains advanced methods to extend geometry building and
    selection within MAPDL.

    """

    def __init__(self, mapdl):
        from ansys.mapdl.core.mapdl import _MapdlCore

        if not isinstance(mapdl, _MapdlCore):
            raise TypeError("Must be initialized using a MAPDL class")

        self._mapdl = mapdl
        self._keypoints_cache = None
        self._lines_cache = None
        self._log = self._mapdl._log

    def _set_log_level(self, level):
        return self._mapdl.set_log_level(level)

    def _load_iges(self):
        """Loads the iges file from MAPDL as a pyiges class"""
        # Lazy import here for speed and stability
        # possible to exclude this import in the future
        try:
            from pyiges import Iges
        except ImportError:
            raise ImportError(
                "Please install pyiges to use this feature with:\n" "pip install pyiges"
            )
        return Iges(self._mapdl._generate_iges())

    def _reset_cache(self):
        self._keypoints_cache = None
        self._lines_cache = None

    @property
    def _keypoints(self):
        """Returns keypoints cache"""
        if self._keypoints_cache is None:
            self._keypoints_cache = self._load_keypoints()
        return self._keypoints_cache

    @property
    def keypoints(self):
        """Keypoint coordinates"""
        return np.asarray(self._keypoints.points)

    @property
    def _lines(self):
        """Returns lines cache"""
        if self._lines_cache is None:
            self._lines_cache = self._load_lines()
        return self._lines_cache

    @property
    def lines(self):
        """Active lines as a pyvista.PolyData"""
        return self._lines

    def areas(self, quality=4, merge=False):
        """List of areas from MAPDL represented as ``pyvista.PolyData``.

        Parameters
        ----------
        quality : int, optional
            quality of the mesh to display.  Varies between 1 (worst)
            to 10 (best).

        merge : bool, optional
            Option to merge areas into a single mesh. Default
            ``False`` to return a list of areas.  When ``True``,
            output will be a single mesh.

        Returns
        -------
        list of pyvista.UnstructuredGrid
            List of ``pyvista.UnstructuredGrid`` meshes representing
            the active surface areas selected by ``ASEL``.  If
            ``merge=True``, areas are returned as a single merged
            UnstructuredGrid.

        Examples
        --------
        Return a list of areas as indiviudal grids

        >>> areas = mapdl.areas(quality=3)
        >>> areab
        [UnstructuredGrid (0x7f14add95040)
          N Cells:	12
          N Points:	20
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	0.000e+00, 0.000e+00
          N Arrays:	4,
        UnstructuredGrid (0x7f14add95ca0)
          N Cells:	12
          N Points:	20
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	5.500e-01, 5.500e-01
          N Arrays:	4,
        ...

        Return a single merged mesh.

        >>> area_mesh = mapdl.areas(quality=3)
        >>> area_mesh
        UnstructuredGrid (0x7f14add95ca0)
          N Cells:	24
          N Points:	30
          X Bounds:	-2.000e+00, 2.000e+00
          Y Bounds:	0.000e+00, 1.974e+00
          Z Bounds:	5.500e-01, 5.500e-01
          N Arrays:	4


        """
        quality = int(quality)
        if quality > 10:
            raise ValueError("``quality`` parameter must be a value between 0 and 10")
        surf = self.generate_surface(11 - quality)
        if merge:
            return surf

        entity_num = surf["entity_num"]
        areas = []
        anums = np.unique(entity_num)
        for anum in anums:
            areas.append(surf.extract_cells(entity_num == anum))

        return areas

    @supress_logging
    @run_as_prep7
    def generate_surface(self, density=4, amin=None, amax=None, ninc=None):
        """
        Generate an all-triangular surface of the active surfaces.

        Parameters
        ----------
        density : int, optional
            APDL smart sizing option.  Ranges from 1 (worst) to 10
            (best).

        amin : int, optional
            Minimum APDL numbered area to select.  See
            ``mapdl.anum`` for available areas.

        amax : int, optional
            Maximum APDL numbered area to select.  See
            ``mapdl.anum`` for available areas.

        ninc : int, optional
            Steps to between amin and amax.
        """
        # store initially selected areas and elements
        with self._mapdl.non_interactive:
            self._mapdl.cm("__tmp_elem__", "ELEM")
            self._mapdl.cm("__tmp_area__", "AREA")
        orig_anum = self.anum

        # reselect from existing selection to mimic APDL behavior
        if amin or amax:
            if amax is None:
                amax = amin

            if amin is None:  # amax is non-zero
                amin = 1

            if ninc is None:
                ninc = ""

            self._mapdl.asel("R", "AREA", vmin=amin, vmax=amax, vinc=ninc)

        # duplicate areas to avoid affecting existing areas
        a_num = int(self._mapdl.get(entity="AREA", item1="NUM", it1num="MAXD"))
        self._mapdl.numstr("AREA", a_num, mute=True)
        self._mapdl.agen(2, "ALL", noelem=1, mute=True)
        a_max = int(self._mapdl.get(entity="AREA", item1="NUM", it1num="MAXD"))

        self._mapdl.asel("S", "AREA", vmin=a_num + 1, vmax=a_max, mute=True)
        # necessary to reset element/area meshing association
        self._mapdl.aatt(mute=True)

        # create a temporary etype
        etype_max = int(self._mapdl.get(entity="ETYP", item1="NUM", it1num="MAX"))
        etype_old = self._mapdl.parameters.type
        etype_tmp = etype_max + 1

        old_routine = self._mapdl.parameters.routine

        self._mapdl.et(etype_tmp, "MESH200", 6, mute=True)
        self._mapdl.shpp("off", mute=True)
        self._mapdl.smrtsize(density, mute=True)
        self._mapdl.type(etype_tmp, mute=True)

        if old_routine != "PREP7":
            self._mapdl.prep7(mute=True)

        # Mesh and get the number of elements per area
        resp = self._mapdl.amesh("all")
        groups = get_elements_per_area(resp)

        self._mapdl.esla("S")
        grid = self._mapdl.mesh._grid.linear_copy()
        pd = pv.PolyData(grid.points, grid.cells, n_faces=grid.n_cells)

        # pd['ansys_node_num'] = grid['ansys_node_num']
        # pd['vtkOriginalPointIds'] = grid['vtkOriginalPointIds']
        # pd.clean(inplace=True)  # OPTIONAL

        # delete all temporary meshes and clean up settings
        self._mapdl.aclear("ALL", mute=True)
        self._mapdl.adele("ALL", kswp=1, mute=True)
        self._mapdl.numstr("AREA", 1, mute=True)
        self._mapdl.type(etype_old, mute=True)
        self._mapdl.etdele(etype_tmp, mute=True)
        self._mapdl.shpp("ON", mute=True)
        self._mapdl.smrtsize("OFF", mute=True)
        self._mapdl.cmsel("S", "__tmp_area__", "AREA", mute=True)
        self._mapdl.cmsel("S", "__tmp_elem__", "ELEM", mute=True)

        # store the area number used for each element
        entity_num = np.empty(grid.n_cells, dtype=np.int32)
        if grid and groups:
            # add anum info
            i = 0
            for index, (anum, nelem) in enumerate(groups):
                # have to use original area numbering here as the
                # duplicated areas numbers are inaccurate
                entity_num[i : i + nelem] = orig_anum[index]
                i += nelem
        else:
            entity_num[:] = 0

        pd["entity_num"] = entity_num
        return pd

    @property
    def n_volu(self):
        """
        Number of volumes currently selected.

        Examples
        --------
        >>> mapdl.n_volu
        1
        """
        return self._item_count("VOLU")

    @property
    def n_area(self):
        """
        Number of areas currently selected.

        Examples
        --------
        >>> mapdl.n_area
        1
        """
        return self._item_count("AREA")

    @property
    def n_line(self):
        """
        Number of lines currently selected.

        Examples
        --------
        >>> mapdl.n_line
        1
        """
        return self._item_count("LINE")

    @property
    def n_keypoint(self):
        """
        Number of keypoints currently selected.

        Examples
        --------
        >>> mapdl.n_keypoint
        1
        """
        return self._item_count("KP")

    @supress_logging
    def _item_count(self, entity):
        """Return item count for a given entity."""
        return int(self._mapdl.get(entity=entity, item1="COUNT"))

    @property
    def knum(self):
        """
        Array of keypoint numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.knum
        array([1, 2, 3, 4, 5, 6, 7, 8], dtype=int32)
        """
        if self._mapdl.geometry.n_keypoint == 0:
            return np.array([], dtype=np.int32)

        return self._mapdl.get_array("KP", item1="KLIST").astype(np.int32)

    @property
    def lnum(self):
        """Array of line numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.lnum
        array([ 1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12], dtype=int32)
        """
        # For clean exit when there is no lines.
        if self._mapdl.geometry.n_line == 0:
            return np.array([], dtype=np.int32)

        # this (weirdly) sometimes fails
        for _ in range(5):
            lnum = self._mapdl.get_array("LINES", item1="LLIST")
            if lnum.size == self.n_line:
                break
        return lnum.astype(np.int32)

    @property
    def anum(self):
        """Array of area numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.anum
        array([1, 2, 3, 4, 5, 6], dtype=int32)
        """
        # Clean exit
        if self._mapdl.geometry.n_area == 0:
            return np.array([], dtype=np.int32)

        return self._mapdl.get_array("AREA", item1="ALIST").astype(np.int32)

    @property
    def vnum(self):
        """Array of volume numbers.

        Examples
        --------
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.vnum
        array([1], dtype=int32)
        """
        if self._mapdl.geometry.n_volu == 0:
            return np.array([], dtype=np.int32)

        return self._mapdl.get_array("VOLU", item1="VLIST").astype(np.int32)

    @supress_logging
    def _load_lines(self):
        """Load lines from MAPDL using IGES"""
        # ignore volumes
        self._mapdl.cm("__tmp_volu__", "VOLU", mute=True)
        self._mapdl.cm("__tmp_line__", "LINE", mute=True)
        self._mapdl.cm("__tmp_area__", "AREA", mute=True)
        self._mapdl.cm("__tmp_keyp__", "KP", mute=True)
        self._mapdl.ksel("ALL", mute=True)
        self._mapdl.lsel("ALL", mute=True)
        self._mapdl.asel("ALL", mute=True)
        self._mapdl.vsel("NONE", mute=True)

        iges = self._load_iges()

        self._mapdl.cmsel("S", "__tmp_volu__", "VOLU", mute=True)
        self._mapdl.cmsel("S", "__tmp_area__", "AREA", mute=True)
        self._mapdl.cmsel("S", "__tmp_line__", "LINE", mute=True)
        self._mapdl.cmsel("S", "__tmp_keyp__", "KP", mute=True)

        selected_lnum = self.lnum
        lines = []
        entity_nums = []
        for bspline in iges.bsplines():
            # allow only 10001 as others appear to be construction entities
            if bspline.d["status_number"] in [1, 10001]:
                entity_num = int(bspline.d["entity_subs_num"])
                if entity_num not in entity_nums and entity_num in selected_lnum:
                    entity_nums.append(entity_num)
                    line = bspline.to_vtk()
                    line.cell_data["entity_num"] = entity_num
                    lines.append(line)

        entities = iges.lines() + iges.circular_arcs()
        for line in entities:
            if line.d["status_number"] == 1:
                entity_num = int(line.d["entity_subs_num"])
                if entity_num not in entity_nums and entity_num in selected_lnum:
                    entity_nums.append(entity_num)
                    line = line.to_vtk(resolution=100)
                    line.cell_data["entity_num"] = entity_num
                    lines.append(line)

        if lines:
            lines = merge_polydata(lines)
            lines["entity_num"] = lines["entity_num"].astype(np.int32)
        else:
            lines = pv.PolyData()

        return lines

    def _load_keypoints(self):
        """Load keypoints from MAPDL using IGES"""
        # write only keypoints
        self._mapdl.cm("__tmp_volu__", "VOLU", mute=True)
        self._mapdl.cm("__tmp_area__", "AREA", mute=True)
        self._mapdl.cm("__tmp_line__", "LINE", mute=True)
        self._mapdl.vsel("NONE", mute=True)
        self._mapdl.asel("NONE", mute=True)
        self._mapdl.lsel("NONE", mute=True)

        iges = self._load_iges()

        self._mapdl.cmsel("S", "__tmp_volu__", "VOLU", mute=True)
        self._mapdl.cmsel("S", "__tmp_area__", "AREA", mute=True)
        self._mapdl.cmsel("S", "__tmp_line__", "LINE", mute=True)

        keypoints = []
        kp_num = []
        for kp in iges.points():
            keypoints.append([kp.x, kp.y, kp.z])
            kp_num.append(int(kp.d["entity_subs_num"]))

        # self._kp_num = np.array(self._kp_num)
        keypoints_pd = pv.PolyData(keypoints)
        keypoints_pd["entity_num"] = kp_num
        return keypoints_pd

    def __str__(self):
        """Current geometry info"""
        info = "MAPDL Selected Geometry\n"
        info += "Keypoints:  %d\n" % self.n_keypoint
        info += "Lines:      %d\n" % self.n_line
        info += "Areas:      %d\n" % self.n_area
        info += "Volumes:    %d\n" % self.n_volu
        return info

    def keypoint_select(self, items, sel_type="S", return_selected=False):
        """Select keypoints using a sequence of items.

        Parameters
        ----------
        items : sequence or None
            List, range, or sequence of integers of the keypoints you want
            to select.  If ``None`` or ``'NONE'``, no keypoints will be
            selected.  If 'ALL', selects all keypoints.

        sel_type : str, optional
            Selection type.  May be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        return_selected : bool, optional
            Return the keypoint numbers selected.  Optional, and can be
            disabled for performance.  Default ``False``.

        Returns
        -------
        list
            List of keypoint numbers if ``return_selected=True``.

        Examples
        --------
        Create a new selection of keypoints [1, 5, 10]

        >>> mapdl.geometry.keypoint_select([1, 5, 10])

        Create a new selection of keypoints from 1 through 20

        >>> mapdl.geometry.keypoint_select(range(1, 21))

        Unselect keypoints 1 through 20

        >>> mapdl.geometry.keypoint_select(range(1, 21), sel_type='U')

        Append to an existing selection of keypoints

        >>> mapdl.geometry.keypoint_select([1, 2, 3], sel_type='A')

        Reselect from the existing selection of keypoints

        >>> mapdl.geometry.keypoint_select([3, 4, 5], sel_type='R')

        Select no keypoints

        >>> mapdl.geometry.keypoint_select(None)

        Select all keypoints

        >>> mapdl.geometry.keypoint_select('ALL')

        """
        if isinstance(items, str):
            items = items.upper()

        # special cases
        if items is None or items == "NONE":
            self._mapdl.ksel("NONE")
            return

        if items == "ALL":
            self._mapdl.ksel("ALL")
            if return_selected:
                return self.knum
            return

        self._select_items(items, "KP", sel_type)

        if return_selected:
            return self.knum

    def line_select(self, items, sel_type="S", return_selected=False):
        """Select lines using a sequence of items.

        Parameters
        ----------
        items : sequence or None
            List, range, or sequence of integers of the lines you want
            to select.  If ``None`` or ``'NONE'``, no lines will be
            selected.  If 'ALL', selects all lines.

        sel_type : str, optional
            Selection type.  May be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        return_selected : bool, optional
            Return the line numbers selected.  Optional, and can be
            disabled for performance.  Default ``False``.

        Returns
        -------
        list
            List of the selected lines if ``return_selected=True``.

        Examples
        --------
        Create a new selection of lines [1, 5, 10]

        >>> mapdl.geometry.line_select([1, 5, 10])

        Create a new selection of lines from 1 through 20

        >>> mapdl.geometry.line_select(range(1, 21))

        Unselect lines 1 through 20

        >>> mapdl.geometry.line_select(range(1, 21), sel_type='U')

        Append to an existing selection of lines

        >>> mapdl.geometry.line_select([1, 2, 3], sel_type='A')

        Reselect from the existing selection of lines

        >>> mapdl.geometry.line_select([3, 4, 5], sel_type='R')

        Select no lines

        >>> mapdl.geometry.line_select(None)

        Select all lines

        >>> mapdl.geometry.line_select('ALL')

        """
        if isinstance(items, str):
            items = items.upper()

        # special cases
        if items is None or items == "NONE":
            self._mapdl.lsel("NONE")
            return

        if items == "ALL":
            self._mapdl.lsel("ALL")
            if return_selected:
                return self.lnum
            return

        self._select_items(items, "LINE", sel_type)

        if return_selected:
            return self.lnum

    def area_select(self, items, sel_type="S", return_selected=False):
        """Select areas using a sequence of items.

        Parameters
        ----------
        items : sequence, str, None
            List, range, or sequence of integers of the areas you want
            to select.  If ``None`` or ``'NONE'``, no areas will be
            selected.  If 'ALL', selects all areas.

        sel_type : str, optional
            Selection type.  May be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        return_selected : bool, optional
            Return the area numbers selected.  Optional, and can be
            disabled for performance.  Default ``False``.

        Returns
        -------
        list
            List of the selected areas if ``return_selected=True``.

        Examples
        --------
        Create a new selection of areas [1, 5, 10]

        >>> mapdl.geometry.area_select([1, 5, 10])

        Create a new selection of areas from 1 through 20

        >>> mapdl.geometry.area_select(range(1, 21))

        Unselect areas 1 through 20

        >>> mapdl.geometry.area_select(range(1, 21), sel_type='U')

        Append to an existing selection of areas

        >>> mapdl.geometry.area_select([1, 2, 3], sel_type='A')

        Reselect from the existing selection of areas

        >>> mapdl.geometry.area_select([3, 4, 5], sel_type='R')

        Select no areas

        >>> mapdl.geometry.area_select(None)

        Select all areas

        >>> mapdl.geometry.area_select('ALL')

        """
        if isinstance(items, str):
            items = items.upper()

        # special cases
        if items is None or items == "NONE":
            self._mapdl.asel("NONE")
            return

        if items == "ALL":
            self._mapdl.asel("ALL")
            if return_selected:
                return self.anum
            return

        self._select_items(items, "AREA", sel_type)

        if return_selected:
            return self.anum

    def volume_select(self, items, sel_type="S", return_selected=False):
        """Select volumes using a sequence of items.

        Parameters
        ----------
        items : sequence, str, or None
            List, range, or sequence of integers of the volumes you want
            to select.  If ``None`` or ``'NONE'``, no volumes will be
            selected.  If 'ALL', selects all volumes.

        sel_type : str, optional
            Selection type.  May be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        return_selected : bool, optional
            Return the volume numbers selected.  Optional, and can be
            disabled for performance.  Default ``False``.

        Returns
        -------
        list
            List of the selected volumes if ``return_selected=True``.

        Examples
        --------
        Create a new selection of volumes [1, 5, 10]

        >>> mapdl.geometry.volume_select([1, 5, 10])

        Create a new selection of volumes from 1 through 20

        >>> mapdl.geometry.volume_select(range(1, 21))

        Unselect volumes 1 through 20

        >>> mapdl.geometry.volume_select(range(1, 21), sel_type='U')

        Append to an existing selection of volumes

        >>> mapdl.geometry.volume_select([1, 2, 3], sel_type='A')

        Reselect from the existing selection of volumes

        >>> mapdl.geometry.volume_select([3, 4, 5], sel_type='R')

        Select no volumes

        >>> mapdl.geometry.volume_select(None)

        Select all volumes

        >>> mapdl.geometry.volume_select('ALL')

        """
        if isinstance(items, str):
            items = items.upper()

        # special cases
        if items is None or items == "NONE":
            self._mapdl.vsel("NONE")
            return

        if items == "ALL":
            self._mapdl.vsel("ALL")
            if return_selected:
                return self.vnum
            return

        self._select_items(items, "VOLU", sel_type)

        if return_selected:
            return self.vnum

    def _select_items(self, items, item_type, sel_type):
        """Select items using FLST

        Parameters
        ----------
        areas : sequence
            Sequence of items.

        item_type : str
            Item lookup type.  One of:

               * 'NODE' : node numbers
               * 'ELEM' : element numbers
               * 'KP' : keypoint numbers
               * 'LINE' : line numbers
               * 'AREA' : area numbers
               * 'VOLU' : volume numbers
               * 'TRACE' : trace points
               * 'COORD' : coordinate locations

        sel_type : str, optional
            Selection type.  Must be one of the following:

            * ``'S'``: Select a new set (default)
            * ``'R'``: Reselect a set from the current set.
            * ``'A'``: Additionally select a set and extend the current set.
            * ``'U'``: Unselect a set from the current set.

        """
        if item_type not in FLST_LOOKUP:
            raise KeyError(f'Invalid ``item_type`` "{item_type}"')

        sel_type = sel_type.upper()
        valid_sel_type = ["S", "R", "A", "U"]
        if sel_type not in valid_sel_type:
            raise ValueError(
                f'Invalid ``sel_type`` "{sel_type}"\n\n'
                f"Use one of the following:\n{VALID_TYPE_MSG}"
            )

        # convert to a flat array as it's easier for type checking
        items = np.asarray(items)
        if not np.issubdtype(items.dtype, np.number):
            raise TypeError("Item numbers must be a numeric type")
        items = items.ravel().astype(np.int, copy=False)

        # consider logic for negative values to support ranges.  This
        # is the 'ORDER' option

        # ordered items
        # FLST,5,76,4,ORDE,74
        # FITEM,5,2
        # LSEL, , , ,P51X

        # unordered option
        with self._mapdl.non_interactive:
            self._mapdl.flst(5, items.size, FLST_LOOKUP[item_type])
            for item in items:
                self._mapdl.fitem(5, item)

            if item_type == "NODE":
                self._mapdl.nsel(sel_type, vmin="P51X")
            elif item_type == "ELEM":
                self._mapdl.esel(sel_type, vmin="P51X")
            elif item_type == "KP":
                self._mapdl.ksel(sel_type, vmin="P51X")
            elif item_type == "LINE":
                self._mapdl.lsel(sel_type, vmin="P51X")
            elif item_type == "AREA":
                self._mapdl.asel(sel_type, vmin="P51X")
            elif item_type == "VOLU":
                self._mapdl.vsel(sel_type, vmin="P51X")
            else:
                raise ValueError(f'Unable to select "{item_type}"')



from .core import _QueryExecution


class _ScalarQueries(_QueryExecution):
    _mapdl = None

    def temp(self, n: int) -> float:
        """Returns temperature at node ``n``.

        Temperature at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Temperature

        Examples
        --------
        We create a block of solid material, and constrain the
        temperature DOF of all its nodes to a uniform value, and then solve.

        Then we can use ``queries.temp`` to get the temperature
        at the first node of the solved model.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> mapdl.clear()
        >>> mapdl.prep7()
        >>> mapdl.mp("kxx", 1, 45)
        >>> mapdl.et(1, 'SOLID70')
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.esize(0.5)
        >>> mapdl.vmesh(1)
        >>> mapdl.d("all", "temp", 5)
        >>> mapdl.slashsolu()
        >>> mapdl.solve()
        >>> mapdl.queries.temp(1)
        5.0
        """
        return self._run_query(f"TEMP({n})", integer=False)

    def pres(self, n: int) -> float:
        """Returns pressure at node ``n``.

        Pressure at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Pressure

        Examples
        --------
        We create a block of solid material, and constrain the
        pressure DOF of all its nodes to a uniform value, and then solve.

        Then we can use ``queries.pres`` to get the pressure
        at the first node of the solved model.

        >>> mapdl.clear()
        >>> mapdl.prep7()
        >>> mapdl.mp("ex", 1, 1)
        >>> mapdl.et(1, 'CPT215')
        >>> mapdl.keyopt(1, 12, 1)
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.esize(0.5)
        >>> mapdl.vmesh(1)
        >>> mapdl.d("all", "pres", 5)
        >>> mapdl.d("all", "ux", 0, lab2="uy", lab3="uz")
        >>> mapdl.run("/SOLU")
        >>> mapdl.solve()
        >>> mapdl.queries.pres(1)
        5.0
        """
        return self._run_query(f"PRES({n})", integer=False)

    def volt(self, n: int) -> float:
        """Returns electric potential at node ``n``.

        Electric potential at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Electric potential

        Examples
        --------
        We create a block of solid material, and constrain the
        volt DOF of all its nodes to a uniform value, and then solve.

        Then we can use ``queries.volt`` to get the Electric Potential
        at the first node of the solved model.

        >>> mapdl.clear()
        >>> mapdl.prep7()
        >>> mapdl.mp("perx", 1, 1)
        >>> mapdl.et(1, 'SOLID122')
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.esize(0.5)
        >>> mapdl.vmesh(1)
        >>> mapdl.d("all", "volt", 5)
        >>> mapdl.run("/SOLU")
        >>> mapdl.solve()
        >>> mapdl.queries.volt(1)
        5.0
        """
        return self._run_query(f"VOLT({n})", integer=False)

    def mag(self, n: int) -> float:
        """Returns magnetic scalar potential at node ``n``.

        Magnetic scalar potential at node ``n``.

        Parameters
        ----------
        n : int
            Node number

        Returns
        -------
        float
            Magnetic scalar potential

        Examples
        --------
        We create a block of solid material, and constrain the
        mag DOF of all its nodes to a uniform value, and then solve.

        Then we can use ``queries.mag`` to get the Magnetic Scalar
        Potential at the first node of the solved model.

        >>> mapdl.clear()
        >>> mapdl.prep7()
        >>> mapdl.mp("murx", 1, 1)
        >>> mapdl.et(1, 'SOLID96')
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.esize(0.5)
        >>> mapdl.vmesh(1)
        >>> mapdl.d("all", "mag", 5)
        >>> mapdl.run("/SOLU")
        >>> mapdl.solve()
        >>> mapdl.queries.mag(1)
        5.0
        """
        return self._run_query(f"MAG({n})", integer=False)



"""Version of ansys-mapdl-core module.

On the ``main`` branch, use 'dev0' to denote a development version.
For example:

# major, minor, patch
version_info = 0, 58, 'dev0'

"""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata

# Read from the pyproject.toml
# major, minor, patch
__version__ = importlib_metadata.version("ansys-mapdl-core")



class Ocean:
    def ocdata(self, val1="", val2="", val3="", val14="", **kwargs):
        """Defines an ocean load using non-table data.

        APDL Command: OCDATA

        Parameters
        ----------
        val1, val2, val3, . . . , val14
            Values describing the basic ocean load or a wave condition.

        Notes
        -----
        The OCDATA command specifies non-table data that defines the ocean
        load, such as the depth of the ocean to the mud line, the ratio of
        added mass over added mass for a circular cross section, or the wave
        type to apply. The terms VAL1, VAL2, etc. are specialized according to
        the input set required for the given ocean load.

        The program interprets the data input via the OCDATA command within the
        context of the most recently issued OCTYPE command.

        Input values in the order indicated.

        This command is also valid in PREP7.

        You can define the following ocean data types:

        For a better understanding of how to set up a basic ocean type, see
        Figure: 5:: Basic Ocean Data Type Components .

        DEPTH -- The depth of the ocean (that is, the distance between the mean
        sea level and the mud line). The water surface is assumed to be level
        in the XY plane, with Z being positive upwards. This value is required
        and must be positive.

        MATOC -- The material number of the ocean. This value is required and
        is used to input the required density. It is also used to input the
        viscosity if the Reynolds number is used (OCTABLE).

        KFLOOD -- The inside-outside fluid-interaction key:

        For beam subtype CTUBE and HREC used with BEAM188 or BEAM189 and ocean
        loading, KFLOOD is always set to 1.

        Cay -- The ratio of added mass of the external fluid over the mass of
        the fluid displaced by the element cross section in the y direction
        (normal). The added mass represents the mass of the external fluid
        (ocean water) that moves with the pipe, beam, or link element when the
        element moves in the element y direction during a dynamic analysis.

        If no value is specified, and the coefficient of inertia CMy is not
        specified (OCTABLE), both values default to 0.0.

        If no value is specified, but CMy is specified, this value defaults to
        Cay = CMy - 1.0.

        If this value should be 0.0, enter 0.0.

        Caz -- The ratio of added mass of the external fluid over the mass of a
        cross section in the element z direction (normal). The added mass
        represents the mass of the external fluid (ocean water) that moves with
        the pipe, beam, or link element when the element moves in the element z
        direction during a dynamic analysis.

        If no value is specified, and Cay is specified, this value defaults to
        Cay.

        If no value is specified, and the coefficient of inertia CMz is not
        specified (OCTABLE), both values default to 0.0.

        If no value is specified, but CMz is specified, this value defaults to
        Cay = CMz - 1.0.

        If this value should be 0.0, enter 0.0.

        Cb -- The ratio of buoyancy force used over buoyancy force based on the
        outside diameter and water density. Accept the default value in most
        cases. Adjust this option only when you must account for additional
        hardware (such as a control valve) attached to the pipe exterior. A
        non-default value may lead to small non-physical inconsistencies;
        testing is therefore recommended for non-default values.

        If no value is specified, this value defaults to 1.0.

        If this value should be 0.0 (useful when troubleshooting your input),
        enter 0.0.

        Zmsl -- A vertical offset from the global origin to the mean sea level.
        The default value is zero (meaning that the origin is located at the
        mean sea level).

        Two example cases for Zmsl are:

        A structure with its origin on the sea floor (Zmsl = DEPTH).

        A tidal change (tc) above the mean sea level (Zmsl = tc, and DEPTH
        becomes DEPTH + tc)

        Ktable -- The dependency of VAL1 on the OCTABLE command:

        KWAVE -- The incident wave type:

        THETA -- Angle of the wave direction θ from the global Cartesian X axis
        toward the global Cartesian Y axis (in degrees).

        WAVELOC (valid when KWAVE = 0 through 3, and 101+) -- The wave location
        type:

        SPECTRUM (valid when KWAVE = 5 through 7) -- The wave spectrum type:

        KCRC -- The wave-current interaction key.

        Adjustments to the current profile are available via the KCRC constant
        of the water motion table. Typically, these options are used only when
        the wave amplitude is large relative to the water depth, such that
        significant wave-current interaction exists.
        """
        command = f"OCDATA,{val1},{val2},{val3},{val14}"
        return self.run(command, **kwargs)

    def ocdelete(self, datatype="", zonename="", **kwargs):
        """Deletes a previously defined ocean load.

        APDL Command: OCDELETE

        Parameters
        ----------
        datatype
            Ocean data type to delete. Valid values are BASIC, CURRENT, WAVE,
            ZONE, and ALL.

        zonename
            The name of the ocean zone to delete. If no name is specified, all
            defined ocean zones are deleted. Valid only when DataType = ZONE.

        Notes
        -----
        The OCDELETE command deletes previously specified ocean data from the
        database.

        This command is also valid in PREP7.
        """
        command = f"OCDELETE,{datatype},{zonename}"
        return self.run(command, **kwargs)

    def oclist(self, datatype="", zonename="", **kwargs):
        """Summarizes all currently defined ocean loads.

        APDL Command: OCLIST

        Parameters
        ----------
        datatype
            Ocean data type to list. Valid values are BASIC, CURRENT, WAVE,
            ZONE, and ALL.

        zonename
            The name of an ocean zone to list. If no name is specified, all
            defined ocean zones are listed. Valid only when DataType = ZONE.

        Notes
        -----
        The OCLIST command summarizes the ocean properties for all defined
        ocean loads in the current session.

        When this command follows the SOLVE command, certain waves types also
        list the calculated wave length.

        This command is also valid in PREP7.
        """
        command = f"OCLIST,{datatype},{zonename}"
        return self.run(command, **kwargs)

    def ocread(self, fname="", ext="", option="", **kwargs):
        """Reads externally defined ocean data.

        APDL Command: OCREAD

        Parameters
        ----------
        fname
            External ocean data file name (excluding the filename extension)
            and directory path containing the file. For more information, see
            the Notes section.

        ext
            Filename extension (limited to eight characters).

        --
            Reserved field.

        option
            Integer value passed to the userOceanRead subroutine (as iOption)
            for user-defined waves. This value does not apply to the diffracted
            wave type.

        Notes
        -----
        The OCREAD command imports ocean data that has been defined externally
        (for example, via the Hydrodynamic Diffraction System (AQWA)).

        The command operates on the ocean load ID specified via the most
        recently issued OCTYPE command. Issue a separate OCREAD command for
        each ocean load that you want to read into the program.

        Fname is limited to 248 characters, including the directory path. If
        Fname does not include a directory path, the program searches for the
        specified file in the current working directory. An unspecified Fname
        defaults to Jobname.

        For the diffracted wave type (KWAVE = 8 on the OCDATA command), you
        must issue an OCREAD command for the ocean wave ID in order to import
        the hydrodynamic data from the hydrodynamic analysis.

        For more information, see Applying Ocean Loading from a Hydrodynamic
        Analysis in the Advanced Analysis Guide.

        To learn more about creating user-defined waves, see Subroutine
        userPanelHydFor (Calculating Panel Loads Caused by Ocean Loading) in
        the Programmer's Reference.

        This command is also valid in PREP7.
        """
        command = f"OCREAD,{fname},{ext},{option}"
        return self.run(command, **kwargs)

    def octable(
        self, val1="", val2="", val3="", val4="", val5="", val6="", val7="", **kwargs
    ):
        """Defines an ocean load using table data.

        APDL Command: OCTABLE

        Parameters
        ----------
        val1, val2, val3, . . . , val6
            Values describing the basic ocean load, a current condition, or a
            wave condition.

        Notes
        -----
        The OCTABLE specifies table data that defines the ocean load. The terms
        VAL1, VAL2, etc. are specialized according to the input set required
        for the given ocean load.

        The program interprets the data input via the OCTABLE command within
        the context of the most recently issued OCTYPE command.

        There is no limit to the number of data input.

        Input values in the order indicated.

        This command is also valid in PREP7.

        You can define the following ocean data types:

        If the current is constant, only one OCTABLE command is necessary and
        Dep  is not required.

        For waves, the current profile is stretched or compressed linearly up
        to 10 percent.

        The first Dep value (representing the mean sea level) must be zero. The
        last Dep value (representing the mud line) must be equal to the DEPTH
        value input on the OCDATA command.

        The Cartesian Z values used to locate nodes, etc. decrease as one moves
        from the ocean surface to the sea floor, but the Dep values increase.
        See Figure: 5:: Basic Ocean Data Type Components .

        Dep is not affected by changes to Zmsl on the OCDATA command, as that
        value simply relocates the origin.

        When specifying an ocean wave type, issue the OCTABLE command to input
        either wave location data or wave spectrum data.

        Hints for Wave Location Input:

        The TIME command is not used, except perhaps to identify the load case.

        The phase shift (Ps) determines the wave position (that is, the
        point at which the load is to be applied).

        When using the Stokes fifth-order (KWAVE = 2) or stream function (KWAVE
        = 3) wave type, issue only one OCTABLE command.

        The valid range of the order of the stream function (NORDER) is 3
        through 50. If no value is specified, the program determines a value
        automatically.

        When using the diffracted wave type (KWAVE = 8), an OCREAD command is
        also required to read in the hydrodynamic data from the hydrodynamic
        analysis.
        """
        command = f"OCTABLE,{val1},{val2},{val3},{val4},{val5},{val6},{val7}"
        return self.run(command, **kwargs)

    def octype(self, datatype="", name="", **kwargs):
        """Specifies the type of ocean load data to follow.

        APDL Command: OCTYPE

        Parameters
        ----------
        datatype
            The type of ocean data to be input following this command:

            BASIC - The basic ocean load, required for any ocean loading.

            CURR - An optional drift current.

            WAVE - An optional ocean wave state.

        name
            An eight-character name for the ocean load. An ocean name can
            consist of letters and numbers, but cannot contain punctuation,
            special characters, or spaces.

        Notes
        -----
        The OCTYPE command specifies the type of ocean load data to follow
        (basic, current, or wave). Issue this command before defining your
        ocean load data (OCDATA and OCTABLE).

        Ocean loading applies only to current-technology pipe (PIPE288 and
        PIPE289), surface (SURF154), link (LINK180) and beam (BEAM188 and
        BEAM189) elements.

        An ocean current or wave is accessible repeatedly. For example, it is
        not necessary to input an identical current table again just because
        the drag coefficients of the basic input table have changed.

        The following example shows how you can use the basic (DataType =
        BASIC), current (DataType = CURR), and wave (DataType = WAVE) ocean
        data types within the context of a simple input file fragment:
        """
        command = f"OCTYPE,{datatype},{name}"
        return self.run(command, **kwargs)

    def oczone(
        self, zonetype="", zonename="", compnameint="", compnameext="", **kwargs
    ):
        """Specifies the type of ocean zone data to follow.

        APDL Command: OCZONE

        Parameters
        ----------
        zonetype
            The type of ocean zone data to be input following this command:

            COMP - Define by a component.

            ZLOC - Define by Z levels.

            PIP - Associate an internal pipe or pipes with an external pipe.

        zonename
            The ocean zone name. If no name is specified, the program assigns
            one.

        compnameint
            For Zonetype = COMP, the required name of a component.

        compnameext
            For Zonetype = PIP, the required name of an external pipe
            component.

        Notes
        -----
        The OCZONE command specifies the type of ocean zone data to follow
        (component, Z-level, or internal pipes associated with an external
        pipe). An ocean zone is a local space where you can override global
        ocean-loading parameters.

        Names specified for ZoneName, CompNameInt, and CompNameExt can consist
        of up to 32 alphanumeric characters. The name cannot contain
        punctuation, special characters, or spaces.

        For Zonetype = COMP, the zone is defined by a component. Only the
        elements in the component are affected by the local parameters. A
        partial component can be defined as the zone via the Z input on the
        OCTABLE command.

        For Zonetype = ZLOC, the zone is defined by Z levels. Structural
        elements (such as BEAM188, BEAM189, PIPE288, PIPE289, and LINK180) in
        the Z levels are included in the zone.

        For Zonetype = PIP, the zone is prepared for a special configuration of
        pipes. It associates an internal pipe or pipes with an external pipe to
        remove the hydrodynamic effect on the internal pipe. Only hydrostatic
        pressure is applied on the internal pipe.

        This command is also valid in PREP7.

        Figure: 6:: : Ocean Zone Types (Specified via ZoneType)

        Issue this command before defining your ocean load data (OCDATA or
        OCTABLE). Define components before defining a component-type or a pipe-
        type zone (OCZONE,COMP or OCZONE,PIP, respectively).
        """
        command = f"OCZONE,{zonetype},{zonename},{compnameint},{compnameext}"
        return self.run(command, **kwargs)



"""Logging module.

This module supplies a general framework for logging in Pymapdl.  This module is
built upon `logging <https://docs.python.org/3/library/logging.html>`_ library
and it does not intend to replace it rather provide a way to interact between
``logging`` and PyMAPDL.

The loggers used in the module include the name of the instance which
is intended to be unique.  This name is printed in all the active
outputs and it is used to track the different MAPDL instances.


Usage
-----

Global logger
~~~~~~~~~~~~~
There is a global logger named ``pymapdl_global`` which is created at
``ansys.mapdl.core.__init__``.  If you want to use this global logger,
you must call at the top of your module:

.. code:: python

   from ansys.mapdl.core import LOG

You could also rename it to avoid conflicts with other loggers (if any):

.. code:: python

   from ansys.mapdl.core import LOG as logger


It should be noticed that the default logging level of ``LOG`` is ``ERROR``.
To change this and output lower level messages you can use the next snippet:

.. code:: python

   LOG.logger.setLevel('DEBUG')
   LOG.file_handler.setLevel('DEBUG')  # If present.
   LOG.stdout_handler.setLevel('DEBUG')  # If present.


Alternatively:

.. code:: python

   LOG.setLevel('DEBUG')

This way ensures all the handlers are set to the input log level.

By default, this logger does not log to a file. If you wish to do so,
you can add a file handler using:

.. code:: python

   import os
   file_path = os.path.join(os.getcwd(), 'pymapdl.log')
   LOG.log_to_file(file_path)

This sets the logger to be redirected also to that file.  If you wish
to change the characteristics of this global logger from the beginning
of the execution, you must edit the file ``__init__`` in the directory
``ansys.mapdl.core``.

To log using this logger, just call the desired method as a normal logger.

.. code:: python

    >>> import logging
    >>> from ansys.mapdl.core.logging import Logger
    >>> LOG = Logger(level=logging.DEBUG, to_file=False, to_stdout=True)
    >>> LOG.debug('This is LOG debug message.')

    DEBUG -  -  <ipython-input-24-80df150fe31f> - <module> - This is LOG debug message.


Instance Logger
~~~~~~~~~~~~~~~
Every time an instance of :class:`Mapdl <ansys.mapdl.core.mapdl._MapdlCore>` is
created, a logger is created and stored in two places:

* ``_MapdlCore._log``. For backward compatibility.
* ``LOG._instances``. This field is a ``dict`` where the key is the name of the
  created logger.

These instance loggers inheritate the ``pymapdl_global`` output handlers and
logging level unless otherwise specified.  The way this logger works is very
similar to the global logger.  You can add a file handler if you wish using
:func:`log_to_file() <PymapdlCustomAdapter.log_to_file>` or change the log level
using :func:`logger.Logging.setLevel`.

You can use this logger like this:

.. code:: python
    >>> from ansys.mapdl.core import launch_mapdl
    >>> mapdl = launch_mapdl()
    >>> mapdl._log.info('This is a useful message')

    INFO - GRPC_127.0.0.1:50056 -  <ipython-input-19-f09bb2d8785c> - <module> - This is a useful message



Other loggers
~~~~~~~~~~~~~
You can create your own loggers using python ``logging`` library as
you would do in any other script.  There shall no be conflicts between
these loggers.

"""

from copy import copy
from datetime import datetime
import logging
import sys
import weakref

## Default configuration
LOG_LEVEL = logging.DEBUG
FILE_NAME = "pymapdl.log"

# For convenience
DEBUG = logging.DEBUG
INFO = logging.INFO
WARN = logging.WARN
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

## Formatting

STDOUT_MSG_FORMAT = (
    "%(levelname)s - %(instance_name)s -  %(module)s - %(funcName)s - %(message)s"
)

FILE_MSG_FORMAT = STDOUT_MSG_FORMAT

DEFAULT_STDOUT_HEADER = """
LEVEL - INSTANCE NAME - MODULE - FUNCTION - MESSAGE
"""
DEFAULT_FILE_HEADER = DEFAULT_STDOUT_HEADER

NEW_SESSION_HEADER = f"""
===============================================================================
       NEW SESSION - {datetime.now().strftime("%m/%d/%Y, %H:%M:%S")}
==============================================================================="""

string_to_loglevel = {
    "DEBUG": DEBUG,
    "INFO": INFO,
    "WARN": WARN,
    "WARNING": WARN,
    "ERROR": ERROR,
    "CRITICAL": CRITICAL,
}


class PymapdlCustomAdapter(logging.LoggerAdapter):
    """This is key to keep the reference to the MAPDL instance name dynamic.

    If we use the standard approach which is supplying ``extra`` input
    to the logger, we would need to keep inputting MAPDL instances
    every time we do a log.

    Using adapters we just need to specify the MAPDL instance we refer
    to once.
    """

    level = None  # This is maintained for compatibility with ``supress_logging``, but it does nothing.
    file_handler = None
    stdout_handler = None

    def __init__(self, logger, extra=None):
        self.logger = logger
        if extra is not None:
            self.extra = weakref.proxy(extra)
        else:
            self.extra = None
        self.file_handler = logger.file_handler
        self.std_out_handler = logger.std_out_handler

    def process(self, msg, kwargs):
        kwargs["extra"] = {}
        # This are the extra parameters sent to log
        kwargs["extra"][
            "instance_name"
        ] = (
            self.extra.get_name()
        )  # here self.extra is the argument pass to the log records.
        return msg, kwargs

    def log_to_file(self, filename=FILE_NAME, level=LOG_LEVEL):
        """Add file handler to logger.

        Parameters
        ----------
        filename : str, optional
            Name of the file where the logs are recorded. By default FILE_NAME
        level : str, optional
            Level of logging. E.x. 'DEBUG'. By default LOG_LEVEL
        """

        self.logger = addfile_handler(
            self.logger, filename=filename, level=level, write_headers=True
        )
        self.file_handler = self.logger.file_handler

    def log_to_stdout(self, level=LOG_LEVEL):
        """Add standard output handler to the logger.

        Parameters
        ----------
        level : str, optional
            Level of logging record. By default LOG_LEVEL
        """
        if self.std_out_handler:
            raise Exception("Stdout logger already defined.")

        self.logger = add_stdout_handler(self.logger, level=level)
        self.std_out_handler = self.logger.std_out_handler

    def setLevel(self, level="DEBUG"):
        """Change the log level of the object and the attached handlers."""
        self.logger.setLevel(level)
        for each_handler in self.logger.handlers:
            each_handler.setLevel(level)
        self.level = level


class PymapdlPercentStyle(logging.PercentStyle):
    def __init__(self, fmt, *, defaults=None):
        self._fmt = fmt or self.default_format
        self._defaults = defaults

    def _format(self, record):
        defaults = self._defaults
        if defaults:
            values = defaults | record.__dict__
        else:
            values = record.__dict__

        # We can do here any changes we want in record, for example adding a key.

        # We could create an if here if we want conditional formatting, and even
        # change the record.__dict__.
        # Since now we don't want to create conditional fields, it is fine to keep
        # the same MSG_FORMAT for all of them.

        # For the case of logging exceptions to the logger.
        values.setdefault("instance_name", "")

        return STDOUT_MSG_FORMAT % values


class PymapdlFormatter(logging.Formatter):
    """Customized ``Formatter`` class used to overwrite the defaults format styles."""

    def __init__(
        self,
        fmt=STDOUT_MSG_FORMAT,
        datefmt=None,
        style="%",
        validate=True,
        defaults=None,
    ):
        if sys.version_info[1] < 8:
            super().__init__(fmt, datefmt, style)
        else:
            # 3.8: The validate parameter was added
            super().__init__(fmt, datefmt, style, validate)
        self._style = PymapdlPercentStyle(fmt, defaults=defaults)  # overwriting


class InstanceFilter(logging.Filter):
    """Ensures that instance_name record always exists."""

    def filter(self, record):
        if not hasattr(record, "instance_name"):
            record.instance_name = ""
        return True


class Logger:
    """Logger used for each Pymapdl session.

    This class allows you to add handlers to the logger to output to a file or
    standard output.

    Parameters
    ----------
    level : int, optional
        Logging level to filter the message severity allowed in the logger.
        The default is ``logging.DEBUG``.
    to_filet : bool, optional
        Write log messages to a file. The default is ``False``.
    to_stdout : bool, optional
        Write log messages into the standard output. The default is
        ``True``.
    filename : str, optional
        Name of the file where log messages are written to.
        The default is ``None``.

    Examples
    --------
    Demonstrate logger usage from an instance mapdl. This is automatically
    created when creating an Mapdl instance.

    >>> from ansys.mapdl.core import launch_mapdl
    >>> mapdl = launch_mapdl(loglevel='DEBUG')
    >>> mapdl._log.info('This is a useful message')
    INFO -  -  <ipython-input-24-80df150fe31f> - <module> - This is LOG debug message.

    Import the global pymapdl logger and add a file output handler.

    >>> import os
    >>> from ansys.mapdl.core import LOG
    >>> file_path = os.path.join(os.getcwd(), 'pymapdl.log')
    >>> LOG.log_to_file(file_path)

    """

    file_handler = None
    std_out_handler = None
    _level = logging.DEBUG
    _instances = {}

    def __init__(
        self, level=logging.DEBUG, to_file=False, to_stdout=True, filename=FILE_NAME
    ):
        """Customized logger class for Pymapdl.

        Parameters
        ----------
        level : str, optional
            Level of logging as defined in the package ``logging``. By default 'DEBUG'.
        to_file : bool, optional
            To record the logs in a file, by default ``False``.
        to_stdout : bool, optional
            To output the logs to the standard output, which is the
            command line. By default ``True``.
        filename : str, optional
            Name of the output file. By default ``pymapdl.log``.
        """

        # create default main logger
        self.logger = logging.getLogger("pymapdl_global")
        self.logger.addFilter(InstanceFilter())
        self.logger.setLevel(level)
        self.logger.propagate = True
        self.level = self.logger.level  # TODO: TO REMOVE

        # Writing logging methods.
        self.debug = self.logger.debug
        self.info = self.logger.info
        self.warning = self.logger.warning
        self.error = self.logger.error
        self.critical = self.logger.critical
        self.log = self.logger.log

        if to_file or filename != FILE_NAME:
            # We record to file
            self.log_to_file(filename=filename, level=level)

        if to_stdout:
            self.log_to_stdout(level=level)

        # Using logger to record unhandled exceptions
        self.add_handling_uncaught_expections(self.logger)

    def log_to_file(self, filename=FILE_NAME, level=LOG_LEVEL):
        """Add file handler to logger.

        Parameters
        ----------
        filename : str, optional
            Name of the file where the logs are recorded. By default
            ``'pymapdl.log'``.
        level : str, optional
            Level of logging. By default ``'DEBUG'``.

        Examples
        --------
        Write to ``pymapdl.log`` in the current working directory.

        >>> from ansys.mapdl.core import LOG
        >>> import os
        >>> file_path = os.path.join(os.getcwd(), 'pymapdl.log')
        >>> LOG.log_to_file(file_path)

        """

        self = addfile_handler(self, filename=filename, level=level, write_headers=True)

    def log_to_stdout(self, level=LOG_LEVEL):
        """Add standard output handler to the logger.

        Parameters
        ----------
        level : str, optional
            Level of logging record. By default  ``'DEBUG'``.
        """

        self = add_stdout_handler(self, level=level)

    def setLevel(self, level="DEBUG"):
        """Change the log level of the object and the attached handlers."""
        self.logger.setLevel(level)
        for each_handler in self.logger.handlers:
            each_handler.setLevel(level)
        self._level = level

    def _make_child_logger(self, sufix, level):
        """Create a child logger.

        Uses ``getChild`` or copying attributes between ``pymapdl_global``
        logger and the new one.
        """
        logger = logging.getLogger(sufix)
        logger.std_out_handler = None
        logger.file_handler = None

        if self.logger.hasHandlers:
            for each_handler in self.logger.handlers:
                new_handler = copy(each_handler)

                if each_handler == self.file_handler:
                    logger.file_handler = new_handler
                elif each_handler == self.std_out_handler:
                    logger.std_out_handler = new_handler

                if level:
                    # The logger handlers are copied and changed the loglevel is
                    # the specified log level is lower than the one of the
                    # global.
                    if each_handler.level > string_to_loglevel[level.upper()]:
                        new_handler.setLevel(level)

                logger.addHandler(new_handler)

        if level:
            if isinstance(level, str):
                level = string_to_loglevel[level.upper()]
            logger.setLevel(level)

        else:
            logger.setLevel(self.logger.level)

        logger.propagate = True
        return logger

    def add_child_logger(self, sufix, level=None):
        """Add a child logger to the main logger.

        This logger is more general than an instance logger which is designed to
        track the state of the MAPDL instances.

        If the logging level is in the arguments, a new logger with a reference
        to the ``_global`` logger handlers is created instead of a child.

        Parameters
        ----------
        sufix : str
            Name of the logger.
        level : str, optional
            Level of logging

        Returns
        -------
        logging.logger
            Logger class.
        """
        name = self.logger.name + "." + sufix
        self._instances[name] = self._make_child_logger(self, name, level)
        return self._instances[name]

    def _add_mapdl_instance_logger(self, name, mapdl_instance, level):
        if isinstance(name, str):
            instance_logger = PymapdlCustomAdapter(
                self._make_child_logger(name, level), mapdl_instance
            )
        elif isinstance(name, None):
            instance_logger = PymapdlCustomAdapter(
                self._make_child_logger("NO_NAMED_YET", level), mapdl_instance
            )
        else:
            raise ValueError("You can only input 'str' classes to this method.")

        return instance_logger

    def add_instance_logger(self, name, mapdl_instance, level=None):
        """Create a logger for a MAPDL instance.

        The MAPDL instance logger is a logger with an adapter which add the
        contextual information such as MAPDL instance name. This logger is
        returned and you can use it to log events as a normal logger. It is also
        stored in the ``_instances`` field.

        Parameters
        ----------
        name : str
            Name for the new logger
        mapdl_instance : ansys.mapdl.core.mapdl._MapdlCore
            Mapdl instance object. This should contain the attribute ``name``.

        Returns
        -------
        ansys.mapdl.core.logging.PymapdlCustomAdapter
            Logger adapter customized to add MAPDL information to the
            logs.  You can use this class to log events in the same
            way you would with the logger class.

        Raises
        ------
        Exception
            You can only input strings as ``name`` to this method.
        """
        count_ = 0
        new_name = name
        while new_name in logging.root.manager.__dict__.keys():
            count_ += 1
            new_name = name + "_" + str(count_)

        self._instances[new_name] = self._add_mapdl_instance_logger(
            new_name, mapdl_instance, level
        )
        return self._instances[new_name]

    def __getitem__(self, key):
        if key in self._instances.keys():
            return self._instances[key]
        else:
            raise KeyError(f"There is no instances with name {key}")

    def add_handling_uncaught_expections(self, logger):
        """This just redirect the output of an exception to the logger."""

        def handle_exception(exc_type, exc_value, exc_traceback):
            if issubclass(exc_type, KeyboardInterrupt):
                sys.__excepthook__(exc_type, exc_value, exc_traceback)
                return
            logger.critical(
                "Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback)
            )

        sys.excepthook = handle_exception


def addfile_handler(logger, filename=FILE_NAME, level=LOG_LEVEL, write_headers=False):
    """Add a file handler to the input.

    Parameters
    ----------
    logger : logging.Logger or logging.Logger
        Logger where to add the file handler.
    filename : str, optional
        Name of the output file. By default FILE_NAME
    level : str, optional
        Level of log recording. By default LOG_LEVEL
    write_headers : bool, optional
        Record the headers to the file. By default False

    Returns
    -------
    logger
        Return the logger or Logger object.
    """

    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(level)
    file_handler.setFormatter(logging.Formatter(FILE_MSG_FORMAT))

    if isinstance(logger, Logger):
        logger.file_handler = file_handler
        logger.logger.addHandler(file_handler)

    elif isinstance(logger, logging.Logger):
        logger.file_handler = file_handler
        logger.addHandler(file_handler)

    if write_headers:
        file_handler.stream.write(NEW_SESSION_HEADER)
        file_handler.stream.write(DEFAULT_FILE_HEADER)

    return logger


def add_stdout_handler(logger, level=LOG_LEVEL, write_headers=False):
    """Add a file handler to the logger.

    Parameters
    ----------
    logger : logging.Logger or logging.Logger
        Logger where to add the file handler.
    level : str, optional
        Level of log recording. By default ``logging.DEBUG``.
    write_headers : bool, optional
        Record the headers to the file. By default ``False``.

    Returns
    -------
    logger
        The logger or Logger object.
    """

    std_out_handler = logging.StreamHandler()
    std_out_handler.setLevel(level)
    std_out_handler.setFormatter(PymapdlFormatter(STDOUT_MSG_FORMAT))

    if isinstance(logger, Logger):
        logger.std_out_handler = std_out_handler
        logger.logger.addHandler(std_out_handler)

    elif isinstance(logger, logging.Logger):
        logger.addHandler(std_out_handler)

    if write_headers:
        std_out_handler.stream.write(DEFAULT_STDOUT_HEADER)

    return logger



class ProcessControls:
    def wait(self, dtime="", **kwargs):
        """APDL Command: /WAIT

        Causes a delay before the reading of the next command.

        Parameters
        ----------
        dtime
            Time delay (in seconds). Maximum time delay is 59 seconds.

        Notes
        -----
        You should consider using ``time.sleep(dtime)``

        The command following the /WAIT will not be processed until the
        specified wait time increment has elapsed.  Useful when reading from a
        prepared input file to cause a pause, for example, after a display
        command so that the display can be reviewed for a period of time.
        Another "wait" feature is available via the ``*ASK`` command.

        This command is valid in any processor.
        """
        command = f"/WAIT,{dtime}"
        return self.run(command, **kwargs)



class GeneralRadiation:
    def stef(self, value="", **kwargs):
        """Specifies Stefan-Boltzmann radiation constant.

        APDL Command: STEF

        Parameters
        ----------
        value
            Stefan-Boltzmann constant (defaults to 0.119E-10 Btu/hr/in2/ °R4).

        Notes
        -----
        You can use this command in the general preprocessor (PREP7) and in the
        Solution processor to specify the Stefan-Boltzmann constant in analyses
        using the radiation matrix method or the radiosity solver to model
        radiation.
        """
        command = f"STEF,{value}"
        return self.run(command, **kwargs)

    def toffst(self, value="", **kwargs):
        """Specifies the temperature offset from absolute zero to zero.

        APDL Command: TOFFST

        Parameters
        ----------
        value
            Degrees between absolute zero and zero of temperature system used
            (should be positive).

        Notes
        -----
        Specifies the difference (in degrees) between absolute zero and the
        zero of the temperature system used.  Absolute temperature values are
        required in evaluating certain expressions, such as for creep,
        swelling, radiation heat transfer, MASS71, etc.  (The offset
        temperature is not used in evaluating emissivity.) Examples are 460°
        for the Fahrenheit system and 273° for the Celsius system.  The offset
        temperature is internally included in the element calculations and does
        not affect the temperature input or output.  If used in SOLUTION, this
        command is valid only within the first load step.

        This command is also valid in PREP7.
        """
        command = f"TOFFST,{value}"
        return self.run(command, **kwargs)



class Status:
    def define(self, **kwargs):
        """Specifies "Data definition settings" as the subsequent status topic.

        APDL Command: DEFINE

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"DEFINE,"
        return self.run(command, **kwargs)

    def operate(self, **kwargs):
        """Specifies "Operation data" as the subsequent status topic.

        APDL Command: OPERATE

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"OPERATE,"
        return self.run(command, **kwargs)

    def plotting(self, **kwargs):
        """Specifies "Plotting settings" as the subsequent status topic.

        APDL Command: PLOTTING

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"PLOTTING,"
        return self.run(command, **kwargs)

    def print(self, **kwargs):
        """Specifies "Print settings" as the subsequent status topic.

        APDL Command: PRINT

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"PRINT,"
        return self.run(command, **kwargs)



"""Miscellaneous methods not covered in the documentation."""


class Misc:
    def verify(self, case="", level="", **kwargs):
        """Enter the verification run mode.

        .. note::
           This command is only valid at the ``/BEGIN`` level, obtained
           with ``mapdl.finish()``.

        Parameters
        ----------
        case : str, optional
            Optional title of the verification manual file.  Also accepts
            ``'OFF'`` to disable the verification run mode.

        level : int, optional
            Verification level ranging from 1 to 6 defaulting to 4.

        Returns
        -------
        str
            Command output.

        Examples
        --------
        Enter the verification routine with the default option.

        >>> mapdl.finish()
        >>> mapdl.verify('VM1')
        '*** VERIFICATION RUN - CASE VM1                              ***  OPTION=  4'

        """
        return self.run(f"/VERIFY,{case},{level}", **kwargs)



"""These DATABASE commands are used to select subsets of database
entities for further operations.
"""

from typing import Optional, Union

from ansys.mapdl.core.mapdl_types import MapdlInt


class Selecting:
    def allsel(self, labt="", entity="", **kwargs):
        """Selects all entities with a single command.

        APDL Command: ALLSEL

        Parameters
        ----------
        labt
            Type of selection to be made:

            ALL - Selects all items of the specified entity type and all
                  items of lower entity types (default).

            BELOW - Selects all items directly associated with and below
                    the selected items of the specified entity type.

        entity
            Entity type on which selection is based:

            ALL - All entity types (default).

            VOLU - Volumes.

            AREA - Areas.

            LINE - Lines.

            KP - Keypoints.

            ELEM - Elements.

            NODE - Nodes.

        Notes
        -----
        ALLSEL is a convenience command that allows the user to select all
        items of a specified entity type or to select items associated with the
        selected items of a higher entity.

        An entity hierarchy is used to decide what entities will be available
        in the selection process.  This hierarchy from top to bottom is as
        follows:  volumes, areas, lines, keypoints, elements, and nodes.  The
        hierarchy may also be divided into two branches:  the solid model and
        the finite element model.  The label ALL selects items based on one
        branch only, while BELOW uses the entire entity hierarchy.  For
        example, ALLSEL,ALL,VOLU selects all volumes, areas, lines, and
        keypoints in the data base.  ALLSEL,BELOW,AREA selects all lines
        belonging to the selected areas; all keypoints belonging to those
        lines; all elements belonging to those areas, lines, and keypoints; and
        all nodes belonging to those elements.

        The $ character should not be used after the  ALLSEL  command.

        This command is valid in any processor.

        Examples
        --------
        >>> mapdl.allsel()
        """
        command = "ALLSEL,%s,%s" % (str(labt), str(entity))
        return self.run(command, **kwargs)

    def asll(self, type_="", arkey="", **kwargs):
        """Selects those areas containing the selected lines.

        APDL Command: ASLL

        Parameters
        ----------
        type\_
            Label identifying the type of area select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        arkey
            Specifies whether all contained area lines must be selected [LSEL]:

            0 - Select area if any of its lines are in the selected line set.

            1 - Select area only if all of its lines are in the selected line set.

        Notes
        -----
        This command is valid in any processor.
        """
        command = "ASLL,%s,%s" % (str(type_), str(arkey))
        return self.run(command, **kwargs)

    def asel(
        self, type_="", item="", comp="", vmin="", vmax="", vinc="", kswp="", **kwargs
    ):
        """Selects a subset of areas.

        APDL Command: ASEL

        Parameters
        ----------
        type\_
            Label identifying the type of select:

            S - Select a new set (default)

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

            ALL - Restore the full set.

            NONE - Unselect the full set.

            INVE - Invert the current set (selected becomes unselected and vice versa).

            STAT - Display the current select status.


        The following fields are used only with Type = S, R, A, or U:

        Item
            Label identifying data. Valid item labels are shown in Table 105: ASEL - Valid Item and Component
            Labels (p. 185). Some items also require a component label. If Item = PICK (or simply "P"), graphical
            picking is enabled and all remaining command fields are ignored (valid only in the GUI). Defaults
            to AREA.

        Comp
            Component of the item (if required). Valid component labels are shown in Table 105: ASEL - Valid
            Item and Component Labels (p. 185).

        VMIN
            Minimum value of item range. Ranges are area numbers, coordinate values, attribute numbers, etc.,
            as appropriate for the item. A component name (as specified on the CM (p. 338) command) may
            also be substituted for VMIN (VMAX and VINC are ignored). If Item = MAT, TYPE, REAL, or ESYS
            and if VMIN is positive, the absolute value of Item is compared against the range for selection; if
            VMIN is negative, the signed value of Item is compared. See the ALIST (p. 106) command for a
            discussion of signed attributes.

        VMAX
            Maximum value of item range. VMAX defaults to VMIN.

        VINC
            Value increment within range. Used only with integer ranges (such as for area numbers). Defaults
            to 1. VINC cannot be negative.

        KSWP
            Specifies whether only areas are to be selected:

                - `kswp = 0` - Select areas only.
                - `kswp = 1` - Select areas, as well as keypoints, lines, nodes, and elements associated with selected areas.

            Valid only with Type = S.


        Notes
        -----
        Selects a subset of areas. For example, to select those areas with area
        numbers 1 through 7, use

        >>> mapdl.asel('S','AREA', '', 1, 7)

        The selected subset is then used when the ALL label is entered
        (or implied) on other commands, such as

        >>> mapdl.alist('ALL')

        Only data identified by area number are selected.  Data
        are flagged as selected and unselected; no data are actually deleted
        from the database.

        In a cyclic symmetry analysis, area hot spots can be modified.
        Consequently, the result of an area selection may be different before
        and after the CYCLIC command.

        If Item = ACCA, the command selects only those areas that were created
        by concatenation.  The KSWP field is processed, but the Comp, VMIN,
        VMAX, and VINC fields are ignored.

        This command is valid in any processor.

        For Selects based on non-integer numbers (coordinates, results, etc.),
        items that are within the range VMIN-Toler and VMAX+Toler are selected.
        The default tolerance Toler is based on the relative values of VMIN and
        VMAX as follows:

        If VMIN = VMAX, Toler = 0.005 x VMIN.

        If VMIN = VMAX = 0.0, Toler = 1.0E-6.

        If VMAX ≠ VMIN, Toler = 1.0E-8 x (VMAX-VMIN).

        Use the SELTOL command to override this default and specify Toler
        explicitly.

        Table: 127:: : ASEL - Valid Item and Component Labels

        Examples
        --------
        Select area(s) at location x == 0.  Note that value of seltol is
        used since ``vmin == vmax``.

        >>> mapdl.asel('S', 'LOC', 'X', 0)

        Select areas between ``y == 2`` and ``y == 4``

        >>> mapdl.asel('S', 'LOC', 'Y', 2, 4)

        """
        command = f"ASEL,{type_},{item},{comp},{vmin},{vmax},{vinc},{kswp}"
        return self.run(command, **kwargs)

    def aslv(self, type_="", **kwargs):
        """Selects those areas contained in the selected volumes.

        APDL Command: ASLV

        Parameters
        ----------
        type\_
            Label identifying the type of area select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        Notes
        -----
        This command is valid in any processor.
        """
        return self.run(f"ASLV,{type_}", **kwargs)

    def dofsel(
        self, type_="", dof1="", dof2="", dof3="", dof4="", dof5="", dof6="", **kwargs
    ):
        """Selects a DOF label set for reference by other commands.

        APDL Command: DOFSEL

        Parameters
        ----------
        type\_
            Label identifying the type of select:

            S - Select a new set of labels.

            A - Add labels to the current set.

            U - Unselect (remove) labels from the current set.

            ALL - Restore the full set of labels.

            STAT - Display the current select status.

        dof1, dof2, dof3, . . . , dof6
            Used only with Type = S, A, or U.  Valid structural labels:  UX,
            UY, or UZ (displacements); U (UX, UY, and UZ);  ROTX, ROTY, or ROTZ
            (rotations); ROT (ROTX, ROTY, and ROTZ);  DISP (U and ROT); HDSP
            (Hydrostatic pressure). Valid thermal labels: TEMP, TBOT, TE2, TE3,
            . . ., TTOP (temperature).  Valid acoustic labels:  PRES
            (pressure); UX, UY, or UZ (displacements for FSI coupled elements).
            Valid electric labels:  VOLT (voltage); EMF (electromotive force
            drop); CURR (current).  Valid magnetic labels:  MAG (scalar
            magnetic potential); AX, AY or AZ (vector magnetic potentials); A
            (AX, AY and AZ); CURR (current).  Valid structural force labels:
            FX, FY, or FZ (forces); F (FX, FY, and FZ); MX, MY, or MZ
            (moments); M (MX, MY, and MZ);  FORC (F and M); DVOL (fluid mass
            flow rate).  Valid thermal force labels:  HEAT, HBOT, HE2, HE3, . .
            ., HTOP (heat flow).  Valid fluid flow force labels:  FLOW (fluid
            flow).  Valid electric force labels:  AMPS (current flow); CHRG
            (electric charge).  Valid magnetic force labels:  FLUX (scalar
            magnetic flux); CSGX, CSGY, or CSGZ (magnetic current segments);
            CSG (CSGX, CSGY, and CSGZ). Valid diffusion labels: CONC
            (concentration); RATE (diffusion flow rate).

        Notes
        -----
        Selects a degree of freedom label set for reference by other commands.
        The label set is used on certain commands where ALL is either input in
        the degree of freedom label field or implied.  The active label set has
        no effect on the solution degrees of freedom.  Specified labels which
        are not active in the model (from the ET or DOF command) are ignored.
        As a convenience, a set of force labels corresponding to the degree of
        freedom labels is also selected.  For example, selecting UX also causes
        FX to be selected (and vice versa).  The force label set is used on
        certain commands where ALL is input in the force label field.

        This command is valid in any processor.
        """
        command = "DOFSEL,%s,%s,%s,%s,%s,%s,%s" % (
            str(type_),
            str(dof1),
            str(dof2),
            str(dof3),
            str(dof4),
            str(dof5),
            str(dof6),
        )
        return self.run(command, **kwargs)

    def esel(
        self,
        type_: str = "",
        item: str = "",
        comp: str = "",
        vmin: Union[str, int, float] = "",
        vmax: Union[str, int, float] = "",
        vinc: MapdlInt = "",
        kabs: MapdlInt = "",
        **kwargs,
    ) -> Optional[str]:
        """Selects a subset of elements.

        APDL Command: ESEL

        Parameters
        ----------
        type\_
            Label identifying the type of select:

            - S - Select a new set (default).
            - R - Reselect a set from the current set.
            - A - Additionally select a set and extend the current set.
            - U - Unselect a set from the current set.
            - ALL - Restore the full set.
            - NONE - Unselect the full set.
            - INVE - Invert the current set (selected becomes
              unselected and vice versa).
            - STAT - Display the current select status.

        item
            Label identifying data, see Table 110: ESEL - Valid Item
            and Component Labels. Some items also require a
            component label. Defaults to ELEM. If Item = STRA
            (straightened), elements are selected whose midside nodes
            do not conform to the curved line or non-flat area on
            which they should lie. (Such elements are sometimes
            formed during volume meshing (VMESH) in an attempt to
            avoid excessive element distortion.) You should
            graphically examine any such elements to evaluate their
            possible effect on solution accuracy.

        comp
            Component of the item (if required). Valid component
            labels are shown in Table 110: ESEL - Valid Item and
            Component Labels below.

        vmin
            Minimum value of item range. Ranges are element numbers,
            attribute numbers, load values, or result values
            as appropriate for the item. A component name (as
            specified via the CM command) can also be substituted for
            VMIN (in which case VMAX and VINC are ignored).

        vmax
            Maximum value of item range. VMAX defaults to VMIN for
            input values. For result values, VMAX defaults to infinity if VMIN is
            positive, or to zero if VMIN is negative.

        vinc
            Value increment within range. Used only with integer
            ranges (such as for element and attribute numbers).
            Defaults to 1. VINC cannot be negative.

        kabs
            Absolute value key:

                - `kabs = 0` - Check sign of value during selection.
                - `kabs = 1` - Use absolute value during selection (sign
                  ignored).

        Notes
        -----
        The fields `item`, `comp`, `vmin`, `vmax`, `vinc` and `kabs` are
        used only with `type_` = `"S"`, `"R"`, `"A"`, or `"U"`.

        Selects elements based on values of a labeled item and
        component. For example, to select a new set of elements
        based on element numbers 1
        through 7, use

        >>> mapdl.esel('S', 'ELEM', '', 1, 7)

        The subset is used when the ALL label is entered (or implied)
        on other commands, such as

        >>> mapdl.elist('ALL')

        Only data identified by element number are
        selected. Selected data are internally flagged; no actual
        removal of data from the database occurs. Different element
        subsets cannot be used for different load steps [SOLVE] in a
        /SOLU sequence.  The subset used in the first load step
        will be used for all subsequent load steps regardless of
        subsequent ESEL specifications.

        This command is valid in any processor.

        Elements crossing the named path (see PATH command) will be
        selected. This option is only available in PREP7 and POST1.
        If no geometry data has been mapped to the path (i.e.,
        via PMAP and PDEF commands), the path will assume the default
        mapping option (PMAP,UNIFORM) to map the geometry prior to
        selecting the elements. If an invalid path name is
        given, the ESEL command is ignored (status of selected
        elements is unchanged). If there are no elements crossing the
        path, the ESEL command will return zero elements selected.

        For selections based on non-integer numbers (coordinates,
        results, etc.), items that are within the range VMIN -Toler
        and VMAX + Toler are selected. The default tolerance Toler is
        based on the relative values of VMIN and VMAX as follows:

        If VMIN = VMAX, Toler = 0.005 x VMIN.

        If VMIN = VMAX = 0.0, Toler = 1.0E-6.

        If VMAX ≠ VMIN, Toler = 1.0E-8 x (VMAX - VMIN).

        Use the SELTOL command to override this default and specify
        Toler explicitly.

        Table: 133:: : ESEL - Valid Item and Component Labels

        Examples
        --------
        Select elements using material numbers 2, 4 and 6.

        >>> mapdl.esel('S', 'MAT', '', 2, 6, 2)

        Of these, reselect SOLID185 element types

        >>> mapdl.esel('R', 'ENAME', '', 185)

        Select elements assigned to material property 2

        >>> mapdl.esel('S', 'MAT', '', 2)

        Note, this command is equivalent to:

        >>> mapdl.esel('S', 'MAT', vmin=2)

        """
        command = f"ESEL,{type_},{item},{comp},{vmin},{vmax},{vinc},{kabs}"
        return self.run(command, **kwargs)

    def esla(self, type_: str = "", **kwargs) -> Optional[str]:
        """Selects those elements associated with the selected areas.

        APDL Command: ESLA

        Parameters
        ----------
        type_
            Label identifying the type of element select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        Notes
        -----
        Selects area elements belonging to meshed [AMESH], selected
        [ASEL] areas.

        This command is valid in any processor.
        """
        command = f"ESLA,{type_}"
        return self.run(command, **kwargs)

    def esll(self, type_: str = "", **kwargs) -> Optional[str]:
        """Selects those elements associated with the selected lines.

        APDL Command: ESLL

        Parameters
        ----------
        type_
            Label identifying the type of element select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        Notes
        -----
        Selects line elements belonging to meshed [LMESH], selected
        [LSEL] lines.

        This command is valid in any processor.
        """
        command = f"ESLL,{type_}"
        return self.run(command, **kwargs)

    def esln(
        self, type_: str = "", ekey: MapdlInt = "", nodetype: str = "", **kwargs
    ) -> Optional[str]:
        """Selects those elements attached to the selected nodes.

        APDL Command: ESLN

        Parameters
        ----------
        type_
            Label identifying the type of element selected:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        ekey
            Node set key:

            0 - Select element if any of its nodes are in the
            selected nodal set (default).

            1 - Select element only if all of its nodes are in the
            selected nodal set.

        nodetype
            Label identifying type of nodes to consider when selecting:

            ALL - Select elements considering all of their nodes (
                  default).

            ACTIVE - Select elements considering only their active
                     nodes. An active node is a node
                     that contributes DOFs to the model.

            INACTIVE - Select elements considering only their
                       inactive nodes (such as orientation or
                       radiation nodes).

            CORNER - Select elements considering only their corner
                     nodes.

            MID - Select elements considering only their midside nodes.

        Notes
        -----
        ESLN selects elements which have any (or all EKEY) NodeType
        nodes in the currently-selected set of nodes. Only elements
        having nodes in the currently-selected set can be selected.

        This command is valid in any processor.
        """
        command = f"ESLN,{type_},{ekey},{nodetype}"
        return self.run(command, **kwargs)

    def eslv(self, type_: str = "", **kwargs) -> Optional[str]:
        """Selects elements associated with the selected volumes.

        APDL Command: ESLV

        Parameters
        ----------
        type_
            Label identifying the type of element selected:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        Notes
        -----
        Selects volume elements belonging to meshed [VMESH], selected
        [VSEL]
        volumes.

        This command is valid in any processor.
        """
        command = f"ESLV,{type_}"
        return self.run(command, **kwargs)

    def ksel(
        self, type_="", item="", comp="", vmin="", vmax="", vinc="", kabs="", **kwargs
    ):
        """Selects a subset of keypoints or hard points.

        APDL Command: KSEL

        Parameters
        ----------
        type\_
            Label identifying the type of select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

            ALL - Restore the full set.

            NONE - Unselect the full set.

            INVE - Invert the current set (selected becomes unselected and vice versa).

            STAT - Display the current select status.


        The following fields are used only with Type = S, R, A, or U:

        Item
            Label identifying data. Valid item labels are shown in the table below. Some items also require a
            component label. If Item = PICK (or simply "P"), graphical picking is enabled and all remaining
            command fields are ignored (valid only in the GUI). Defaults to KP.

        Comp
            Component of the item (if required). Valid component labels are shown in the table below.

        VMIN
            Minimum value of item range. Ranges are keypoint numbers, coordinate values, attribute numbers,
            etc., as appropriate for the item. A component name (as specified on the CM (p. 338) command)
            may also be substituted for VMIN (VMAX and VINC are ignored). If Item = MAT, TYPE, REAL, or
            ESYS and if VMIN is positive, the absolute value of Item is compared against the range for selection;
            if VMIN is negative, the signed value of Item is compared. See the KLIST (p. 942) command for a
            discussion of signed attributes.

        VMAX
            Maximum value of item range. VMAX defaults to VMIN.

        VINC
            Value increment within range. Used only with integer ranges (such as for keypoint numbers). Defaults
            to 1. VINC cannot be negative.

        KABS
            Absolute value key:

                - `kabs = 0` - Check sign of value during selection.
                - `kabs = 1` - Use absolute value during selection (sign ignored).

        Notes
        -----
        Selects a subset of keypoints or hard points.  For example, to select a
        new set of keypoints based on keypoint numbers 1 through 7, use

        >>> mapdl.ksel('S', 'KP', '', 1, 7)

        The selected subset is used when the ALL label is
        entered (or implied) on other commands, such as KLIST,ALL.  Only data
        identified by keypoint number are selected.  Data are flagged as
        selected and unselected; no data are actually deleted from the
        database.

        This command is valid in any processor.

        For selections based on non-integer numbers (coordinates, results,
        etc.), items that are within the range VMIN -Toler and VMAX + Toler are
        selected. The default tolerance Toler is based on the relative values
        of VMIN and VMAX as follows:

        If VMIN = VMAX, Toler = 0.005 x VMIN.

        If VMIN = VMAX = 0.0, Toler = 1.0E-6.

         If VMAX ≠ VMIN, Toler = 1.0E-8 x (VMAX - VMIN).

        Use the SELTOL command to override this default and specify Toler
        explicitly.

        Table: 203:: : KSEL - Valid Item and Component Labels

        Examples
        --------
        To select a single keypoint (keypoint 1)

        >>> mapdl.ksel('S', 'KP', '', 1)
        """
        command = "KSEL,%s,%s,%s,%s,%s,%s,%s" % (
            str(type_),
            str(item),
            str(comp),
            str(vmin),
            str(vmax),
            str(vinc),
            str(kabs),
        )
        return self.run(command, **kwargs)

    def ksll(self, type_="", **kwargs):
        """Selects those keypoints contained in the selected lines.

        APDL Command: KSLL

        Parameters
        ----------
        type\_
            Label identifying the type of keypoint select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        Notes
        -----
        This command is valid in any processor.
        """
        command = "KSLL,%s" % (str(type_))
        return self.run(command, **kwargs)

    def ksln(self, type_="", **kwargs):
        """Selects those keypoints associated with the selected nodes.

        APDL Command: KSLN

        Parameters
        ----------
        type\_
            Label identifying the type of keypoint select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        Notes
        -----
        Valid only if the nodes were generated by a meshing operation [KMESH,
        LMESH, AMESH, VMESH] on a solid model that contains the associated
        keypoints.

        This command is valid in any processor.
        """
        command = "KSLN,%s" % (str(type_))
        return self.run(command, **kwargs)

    def lsel(
        self, type_="", item="", comp="", vmin="", vmax="", vinc="", kswp="", **kwargs
    ):
        """Selects a subset of lines.

        APDL Command: LSEL

        Parameters
        ----------
        type\_
            Label identifying the type of select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

            ALL - Restore the full set.

            NONE - Unselect the full set.

            INVE - Invert the current set (selected becomes unselected and vice versa).

            STAT - Display the current select status.


        The following fields are used only with Type = S, R, A, or U:

        Item
            Label identifying data. Valid item labels are shown in the table
            below. Some items also require a component label. If Item = PICK
            (or simply "P"), graphical picking is enabled and all remaining
            command fields are ignored (valid only in the GUI). Defaults to
            LINE.

        Comp
            Component of the item (if required). Valid component labels are
            shown in the table below.

        VMIN
            Minimum value of item range. Ranges are line numbers, coordinate
            values, attribute numbers, etc., as appropriate for the item. If
            VMIN = 0.0, a tolerance of ±1.0E-6 is used, or ±0.005 x VMIN if
            VMIN = VMAX. A component name (as specified on the CM (p. 338)
            command) may also be substituted for VMIN (VMAX and VINC are
            ignored). If Item = MAT, TYPE, REAL, ESYS, or NDIV and if VMIN is
            positive, the absolute value of Item is compared against the range
            for selection; if VMIN is negative, the signed value of Item is
            compared. See the LLIST (p. 1011) command for a discussion of
            signed attributes.

        VMAX
            Maximum value of item range. VMAX defaults to VMIN.

        VINC
            Value increment within range. Used only with integer ranges (such
            as for line numbers). Defaults to 1. VINC cannot be negative.

        KSWP
            Specifies whether only lines are to be selected:

                - ``kswp=0`` - Select lines only.  ``kswp=1`` - Select lines,
                - as well as keypoints, nodes, and elements associated with
                - selected lines. Valid only with ``type="s"``.

        Notes
        -----
        Selects lines based on values of a labeled item and component.  For
        example, to select a new set of lines based on line numbers 1 through
        7, use LSEL,S,LINE,,1,7.  The subset is used when the ALL label is
        entered (or implied) on other commands, such as LLIST,ALL.  Only data
        identified by line number are selected.  Data are flagged as selected
        and unselected; no data are actually deleted from the database.

        If Item = LCCA, the command selects only those lines that were created
        by concatenation.  The KSWP field is processed, but the Comp, VMIN,
        VMAX, and VINC fields are ignored.

        If Item = HPT, the command selects only those lines that contain hard
        points.

        Item = RADIUS is only valid for lines that are circular arcs.

        LSEL is valid in any processor.

        For selections based on non-integer numbers (coordinates, results,
        etc.), items that are within the range VMIN -Toler and VMAX +Toler are
        selected. The default tolerance Toler is based on the relative values
        of VMIN and VMAX as follows:

        If VMIN = VMAX, Toler = 0.005 x VMIN.

        If VMIN = VMAX = 0.0, Toler = 1.0E-6.

        If VMAX ≠ VMIN, Toler = 1.0E-8 x (VMAX - VMIN).

        Use the SELTOL command to override this default and specify Toler
        explicitly.

        """
        command = "LSEL,%s,%s,%s,%s,%s,%s,%s" % (
            str(type_),
            str(item),
            str(comp),
            str(vmin),
            str(vmax),
            str(vinc),
            str(kswp),
        )
        return self.run(command, **kwargs)

    def lsla(self, type_="", **kwargs):
        """Selects those lines contained in the selected areas.

        APDL Command: LSLA

        Parameters
        ----------
        type\_
            Label identifying the type of line select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        Notes
        -----
        This command is valid in any processor.
        """
        command = "LSLA,%s" % (str(type_))
        return self.run(command, **kwargs)

    def lslk(self, type_="", lskey="", **kwargs):
        """Selects those lines containing the selected keypoints.

        APDL Command: LSLK

        Parameters
        ----------
        type\_
            Label identifying the type of line select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        lskey
            Specifies whether all contained line keypoints must be selected
            [KSEL]:

            0 - Select line if any of its keypoints are in the selected keypoint set.

            1 - Select line only if all of its keypoints are in the selected keypoint set.

        Notes
        -----
        This command is valid in any processor.
        """
        command = "LSLK,%s,%s" % (str(type_), str(lskey))
        return self.run(command, **kwargs)

    def nsel(
        self, type_="", item="", comp="", vmin="", vmax="", vinc="", kabs="", **kwargs
    ):
        """Selects a subset of nodes.

        APDL Command: NSEL

        Parameters
        ----------
        type\_
            Label identifying the type of select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

            ALL - Restore the full set.

            NONE - Unselect the full set.

            INVE - Invert the current set (selected becomes unselected and vice versa).

            STAT - Display the current select status.


        The following fields are used only with Type = S, R, A, or U:

        Item
            Label identifying data. Valid item labels are shown in the table below. Some items also require a
            component label. If Item = PICK (or simply "P"), graphical picking is enabled and all remaining
            command fields are ignored (valid only in the GUI). Defaults to NODE.

        Comp
            Component of the item (if required). Valid component labels are shown in the table below.

        VMIN
            Minimum value of item range. Ranges are node numbers, set numbers, coordinate values, load
            values, or result values as appropriate for the item. A component name (as specified on the CM (p. 338)
            command) may also be substituted for VMIN (VMAX and VINC are ignored).

        VMAX
            Maximum value of item range. VMAX defaults to VMIN for input values. For result values, VMAX
            defaults to infinity if VMIN is positive, or to zero if VMIN is negative.

        VINC
            Value increment within range. Used only with integer ranges (such as for node and set numbers).
            Defaults to 1. VINC cannot be negative.

        KABS
            Absolute value key:

            - `kabs = 0` - Check sign of value during selection.
            - `kabs = 1` - Use absolute value during selection (sign ignored)

        Notes
        -----
        Selects a subset of nodes.  For example, to select a new set of nodes
        based on node numbers 1 through 7, do

        >>> mapdl.nsel('S','NODE','',1,7)

        The subset is used when the `'ALL'` label is entered (or implied)
        on other commands, such as

        >>> mapdl.nlist('ALL')

        Only data identified by node number are selected.  Data
        are flagged as selected and unselected; no data are actually deleted
        from the database.

        When selecting nodes by results, the full graphics value is used,
        regardless of whether PowerGraphics is on.

        Solution result data consists of two types, 1) nodal degree of freedom
        --results initially calculated at the nodes (such as displacement,
        temperature, pressure, etc.), and 2) element--results initially
        calculated elsewhere (such as at an element integration point or
        thickness location) and then recalculated at the nodes (such as
        stresses, strains, etc.).  Various element results also depend upon the
        recalculation method and the selected results location [AVPRIN, RSYS,
        FORCE, LAYER and SHELL].

        You must have all the nodes (corner and midside nodes) on the external
        face of the element selected to use Item = EXT.

        This command is valid in any processor.

        For Selects based on non-integer numbers (coordinates, results, etc.),
        items that are within the range VMIN-Toler and VMAX+Toler are selected.
        The default tolerance Toler is based on the relative values of VMIN and
        VMAX as follows:

        If VMIN = VMAX, Toler = 0.005 x VMIN.

        If VMIN = VMAX = 0.0, Toler = 1.0E-6.

         If VMAX ≠ VMIN, Toler = 1.0E-8 x (VMAX-VMIN).

        Use the SELTOL command to override this default and specify Toler
        explicitly.

        Table: 208:: : NSEL - Valid Item and Component Labels

        Table: 209:: : NSEL - Valid Item and Component Labels for Nodal DOF
        Result Values

        Examples
        --------
        Select nodes at `x == 0`, Of these, reselect nodes between ``1 < Y
        < 10``, then finally unselect nodes between ``5 < Y < 6``.

        >>> mapdl.nsel('S', 'LOC', 'X', 0)
        >>> mapdl.nsel('R', 'LOC', 'Y', 1, 10)
        >>> mapdl.nsel('U', 'LOC', 'Y', 5, 6)

        For other coordinate systems activate the coord system first.
        First, change to cylindrical coordinate system, select nodes at
        ``radius == 5``.  Reselect nodes from 0 to 90 degrees.

        >>> mapdl.csys(1)
        >>> mapdl.nsel('S', 'LOC', 'X', 5)
        >>> mapdl.nsel('R', 'LOC', 'Y', 0, 90)

        Note that the labels X, Y, and Z are always used, regardless of which
        coordinate system is activated. They take on different meanings in
        different systems Additionally, angles are always in degrees and
        NOT radians.

        Select elements assigned to material property 2

        >>> mapdl.esel('S', 'MAT', '', 2)

        Select the nodes these elements use

        >>> mapdl.nsle()

        Reselect nodes on the element external surfaces

        >>> mapdl.nsel('R', 'EXT')

        Change to cylindrical coordinates

        >>> mapdl.csys(1)

        Reselect nodes with radius=5

        >>> mapdl.nsel('R', 'LOC', 'X', 5)
        """
        command = "NSEL,%s,%s,%s,%s,%s,%s,%s" % (
            str(type_),
            str(item),
            str(comp),
            str(vmin),
            str(vmax),
            str(vinc),
            str(kabs),
        )
        return self.run(command, **kwargs)

    def nsla(self, type_="", nkey="", **kwargs):
        """Selects those nodes associated with the selected areas.

        APDL Command: NSLA

        Parameters
        ----------
        type\_
            Label identifying the type of node select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        nkey
            Specifies whether only interior area nodes are to be selected:

            0 - Select only nodes interior to selected areas.

            1 - Select all nodes (interior to area, interior to lines, and at keypoints)
                associated with the selected areas.

        Notes
        -----
        Valid only if the nodes were generated by an area meshing operation
        [AMESH, VMESH] on a solid model that contains the selected areas.

        This command is valid in any processor.

        Examples
        --------
        Select area(s) at location x=0.  Note that the selection tolerance
        above and below 0 will depend on the value of ``mapdl.seltol``

        >>> mapdl.asel('S', 'LOC', 'X', 0)

        Select nodes residing on those areas.

        >>> mapdl.nsla('S', 1)
        """
        command = "NSLA,%s,%s" % (str(type_), str(nkey))
        return self.run(command, **kwargs)

    def nsle(self, type_="", nodetype="", num="", **kwargs):
        """Selects those nodes attached to the selected elements.

        APDL Command: NSLE

        Parameters
        ----------
        type\_
            Label identifying the type of node select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        nodetype
            Label identifying type of nodes to consider when selecting:

            ALL - Select all nodes of the selected elements (default).

            ACTIVE - Select only the active nodes. An active node is a node that contributes DOFs to
                     the model.

            INACTIVE - Select only inactive  nodes (such as orientation or radiation).

            CORNER - Select only corner nodes.

            MID - Select only midside nodes.

            POS - Select nodes in position Num.

            FACE - Select nodes on face Num.

        num
            Position or face number for NodeType = POS or FACE.

        Notes
        -----
        NSLE selects NodeType nodes attached to the currently-selected set of
        elements. Only nodes on elements in the currently-selected element set
        can be selected.

        .. note::
            When using degenerate hexahedral elements, NSLE, U,CORNER and
            NSLE,S,MID will not select the same set of nodes because some
            nodes appear as both corner and midside nodes.

        This command is valid in any processor.

        Examples
        --------
        Select elements assigned to material property 2 and then select
        the nodes these elements use.

        >>> mapdl.esel('S', 'MAT', '', 2)
        >>> mapdl.nsle()
        """
        command = "NSLE,%s,%s,%s" % (str(type_), str(nodetype), str(num))
        return self.run(command, **kwargs)

    def nslk(self, type_="", **kwargs):
        """Selects those nodes associated with the selected keypoints.

        APDL Command: NSLK

        Parameters
        ----------
        type\_
            Label identifying the type of node select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        Notes
        -----
        Valid only if the nodes were generated by a keypoint meshing operation
        [KMESH, LMESH, AMESH, VMESH] on a solid model that contains the
        selected keypoints.

        This command is valid in any processor.
        """
        command = "NSLK,%s" % (str(type_))
        return self.run(command, **kwargs)

    def nsll(self, type_="", nkey="", **kwargs):
        """Selects those nodes associated with the selected lines.

        APDL Command: NSLL

        Parameters
        ----------
        type\_
            Label identifying the type of node select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        nkey
            Specifies whether only interior line nodes are to be selected:

            0 - Select only nodes interior to selected lines.

            1 - Select all nodes (interior to line and at keypoints)
                associated with the selected lines.

        Notes
        -----
        Valid only if the nodes were generated by a line meshing
        operation [LMESH, AMESH, VMESH] on a solid model that contains
        the associated lines.

        This command is valid in any processor.
        """
        command = "NSLL,%s,%s" % (str(type_), str(nkey))
        return self.run(command, **kwargs)

    def nslv(self, type_="", nkey="", **kwargs):
        """Selects those nodes associated with the selected volumes.

        APDL Command: NSLV

        Parameters
        ----------
        type\_
            Label identifying the type of node select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        nkey
            Specifies whether only interior volume nodes are to be selected:

            0 - Select only nodes interior to selected volumes.

            1 - Select all nodes (interior to volume, interior to areas, interior to lines, and
                at keypoints) associated with the selected volumes.

        Notes
        -----
        Valid only if the nodes were generated by a volume meshing operation
        [VMESH] on a solid model that contains the selected volumes.

        This command is valid in any processor.
        """
        command = "NSLV,%s,%s" % (str(type_), str(nkey))
        return self.run(command, **kwargs)

    def partsel(self, type_="", pmin="", pmax="", pinc="", **kwargs):
        """Selects a subset of parts in an explicit dynamic analysis.

        APDL Command: PARTSEL

        Parameters
        ----------
        type\_
            Label identifying type of select. Because PARTSEL is a command
            macro, the label must be enclosed in single quotes.

            'S' - Select a new set (default).

            'R' - Reselect a set from the current set.

            'A' - Additionally select a set and extend the current set.

            'U' - Unselect a set from the current set.

            'ALL' - Select all parts.

            'NONE' - Unselect all parts.

            'INVE' - Invert the current selected set.

        Notes
        -----
        PARTSEL invokes an ANSYS macro that selects parts in an explicit
        dynamic analysis. When PARTSEL is executed, an element component is
        automatically created for each existing part. For example, the elements
        that make up PART 1 are grouped into the element component
        _PART1. Each time the PARTSEL command is executed, components
        for unselected parts will be unselected. To plot selected parts,
        choose Utility Menu> Plot> Parts in the GUI or issue the
        command PARTSEL,'PLOT'.

        After selecting parts, if you change the selected set of nodes or
        elements and then plot parts, the nodes and elements associated with
        the previously selected parts (from the last PARTSEL command) will
        become the currently selected set.

        Note:: : A more efficient way to select and plot parts is to use the
        ESEL (with ITEM = PART) and EPLOT commands. We recommend using ESEL
        instead of PARTSEL since PARTSEL will be phased out in a future
        release. Note that the menu path mentioned above for plotting parts
        does not work with the ESEL command; use Utility Menu> Plot> Elements
        instead.

        In an explicit dynamic small restart analysis (EDSTART,2), PARTSEL can
        be used to unselect a part during the solution even if it is referenced
        in some way (such as in a contact definition). (Note that ESEL cannot
        be used for this purpose.) However, in a new analysis or a full restart
        analysis (EDSTART,3), all parts that are used in some type of
        definition must be selected at the time of solution.

        This command is valid in any processor.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = "PARTSEL,%s,%s,%s,%s" % (str(type_), str(pmin), str(pmax), str(pinc))
        return self.run(command, **kwargs)

    def vsel(
        self, type_="", item="", comp="", vmin="", vmax="", vinc="", kswp="", **kwargs
    ):
        """Selects a subset of volumes.

        APDL Command: VSEL

        Parameters
        ----------
        type\_
            Label identifying the type of volume select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

            ALL - Restore the full set.

            NONE - Unselect the full set.

            INVE - Invert the current set (selected becomes unselected and vice versa).

            STAT - Display the current select status.

        item : str, optional
            Label identifying data. Valid item labels are shown in the table
            below. Some items also require a component label. Defaults to
            VOLU.

        comp : str, optional
            Component of the item (if required).

        vmin : int, optional
            Minimum value of item range. Ranges are volume numbers, coordinate
            values, attribute numbers, etc., as appropriate for the item. A
            component name (as specified on the CM (p. 338) command) may also
            be substituted for VMIN (VMAX and VINC are ignored). If Item = MAT,
            TYPE, REAL, or ESYS and if VMIN is positive, the absolute value of
            Item is compared against the range for selection; if VMIN is
            negative, the signed value of Item is compared. See the VLIST
            (p. 2050) command for a discussion of signed attributes.

        vmax : int, optional
            Maximum value of item range. VMAX defaults to VMIN.

        vinc : int, optional
            Value increment within range. Used only with integer ranges (such
            as for volume numbers). Defaults to 1. VINC cannot be negative.

        kswp : int, optional
            Specifies whether only volumes are to be selected:

            - ``kswp=0`` - Select volumes only.
            - ``kswp=1`` - Select volumes, as well as keypoints, lines, areas,
              nodes, and elements associated with selected volumes. Valid only
              with Type = S.

        Notes
        -----
        Selects volumes based on values of a labeled item and component.  For
        example, to select a new set of volumes based on volume numbers 1
        through 7, use VSEL,S,VOLU,,1,7.  The subset is used when the ALL label
        is entered (or implied) on other commands, such as VLIST,ALL.  Only
        data identified by volume number are selected.  Data are flagged as
        selected and unselected; no data are actually deleted from the
        database.

        This command is valid in any processor.

        For Selects based on non-integer numbers (coordinates, results, etc.),
        items that are within the range VMIN-Toler and VMAX+Toler are selected.
        The default tolerance Toler is based on the relative values of VMIN and
        VMAX as follows:

        If VMIN = VMAX, Toler = 0.005 x VMIN.

        If VMIN = VMAX = 0.0, Toler = 1.0E-6.

        If VMAX ≠ VMIN, Toler = 1.0E-8 x (VMAX-VMIN).

        Use the SELTOL command to override this default and specify Toler
        explicitly.

        """
        command = "VSEL,%s,%s,%s,%s,%s,%s,%s" % (
            str(type_),
            str(item),
            str(comp),
            str(vmin),
            str(vmax),
            str(vinc),
            str(kswp),
        )
        return self.run(command, **kwargs)

    def vsla(self, type_="", vlkey="", **kwargs):
        """Selects those volumes containing the selected areas.

        APDL Command: VSLA

        Parameters
        ----------
        type\_
            Label identifying the type of volume select:

            S - Select a new set (default).

            R - Reselect a set from the current set.

            A - Additionally select a set and extend the current set.

            U - Unselect a set from the current set.

        vlkey
            Specifies whether all contained volume areas must be selected
            [ASEL]:

            0 - Select volume if any of its areas are in the selected area set.

            1 - Select volume only if all of its areas are in the selected area set.

        Notes
        -----
        This command is valid in any processor.
        """
        command = "VSLA,%s,%s" % (str(type_), str(vlkey))
        return self.run(command, **kwargs)

    def seltol(self, toler="", **kwargs):
        """Select those volumes containing the selected areas.

        APDL Command: SELTOL

        Parameters
        ----------
        toler
            Tolerance value. If blank or None, restores the default tolerance
            logic.

        Notes
        -----
        For selects based on non-integer numbers (e.g. coordinates,
        results, etc.), items within the range VMIN - Toler and VMAX +
        Toler are selected, where VMIN and VMAX are the range values input
        on the xSEL commands (ASEL, ESEL, KSEL, LSEL, NSEL, and VSEL).

        The default tolerance logic is based on the relative values of
        VMIN and VMAX as follows:

        If ``vmin == vmax``, ``toler = 0.005*vmin``

        If ``vmin == vmax == 0.0``, ``toler = 1.0E-6``.

        If ``vmax != vmin``, ``toler = 1.0E-8 x (vmax - vmin)``.

        This command is typically used when ``vmax - vmin`` is very large so
        that the computed default tolerance is therefore large and the
        xSEL commands selects more than what is desired.

        Toler remains active until respecified by a subsequent seltol
        command. ``seltol()`` resets back to the default toler value.

        Examples
        --------
        Set selection tolarance to 1E-5

        >>> seltol(1E-5)

        """
        if toler:
            cmd = f"SELTOL,{toler}"
        else:
            cmd = "SELTOL"
        return self.run(cmd, **kwargs)
